/* 
 * This file is part of the Modbus IO distribution.
 * Copyright (c) 2020 Martin Wende.
 * 
 * This program is free software: you can redistribute it and/or modify  
 * it under the terms of the GNU General Public License as published by  
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef I_MB_UART_HPP
#define I_MB_UART_HPP

#include "mb_defs.hpp"

//! \brief uart errs from hardware
enum class mb_uart_err {
	SUCCESS 	= 0,		//!< no errors
	ERR_PARITY 	= 1 << 0,	//!< parity error occured
	ERR_OVERRUN = 1 << 1,	//!< overrun occured (better use dma with uart hw timeouts)
	ERR_FRAME	= 1 << 2,	//!< frame error occured (wrong baudrate?)
	ERR_DMA		= 1 << 3,	//!< error in dma occured
};

//! \brief mb uart callback
//! \param[in] buf buffer that was received/send
//! \param[in] len len that was received/send
//! \param[in] err errflags that happened while sending/receiving
//! \param[in] arg userarg given
typedef void (*mb_uart_callb)(uint8_t*, uint32_t, uint32_t, void*);

//! \brief an interface of a modbus uart so that the protocol stack doesnt have to know the hardware
class i_mb_uart {
public:
	
	//! \brief init with uart mode
	//! \param[in] baudrate baudrate to set
	//! \param[in] parity parity to set
	//! \return @ref mb_err
	virtual mb_err init(mb_baudrate baudrate, mb_parmode parity) = 0;
	
	//! \brief check if baudrate is possible before set
	//! \param[in] baudrate baudrate to set
	//! \return @ref mb_err
	virtual mb_err check_baud(mb_baudrate baudrate) = 0;
	
	//! \brief start receiving (again)
	//! \param[out] buf buffer to store received frame to (length is known)
	//! \param[in] len max length of frame to receive
	//! \return @ref mb_err
	virtual mb_err receive(uint8_t *buf, uint16_t len) = 0;
	
	//! \brief start sending
	//! \param[in] buf buffer to send
	//! \param[in] len length of frame to send
	//! \return @ref mb_err
	virtual mb_err send(uint8_t *buf, uint16_t len) = 0;
	
	//! \brief register a callback for receiving
	//! \param[in] callb callback to register (can be NULL)
	//! \param[in] userarg arg to call the callback with
	//! \return @ref mb_err
	virtual mb_err register_callback_rx(mb_uart_callb callb, void *userarg) = 0;
	
	//! \brief register a callback for sending
	//! \param[in] callb callback to register (can be NULL)
	//! \param[in] userarg arg to call the callback with
	//! \return @ref mb_err
	virtual mb_err register_callback_tx(mb_uart_callb callb, void *userarg) = 0;
	
	//! \brief call this if the IRQ happened
	virtual void irq() = 0;
};

#endif//I_MB_UART_HPP