/* 
 * This file is part of the DPS Entkoppler distribution.
 * Copyright (c) 2020 Martin Wende.
 * 
 * This program is free software: you can redistribute it and/or modify  
 * it under the terms of the GNU General Public License as published by  
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef MB_DEFS_H
#define MB_DEFS_H

#include "stdint.h"

#define MB_ADDRESS_DEFAULT	247		// the modbus default address if none is assigned yet

#define MB_FRAME_SIZEMAX 	256		// the maximum size of a MODBUS RTU frame is 256 bytes
#define MB_FRAME_SIZEMIN	4		// min 1 byte Adress, 1 byte function code and 2 byte CRC
#define MB_FRAME_SIZEEXC	5		// size of exception frame

#define MB_MAXSIZE_READ_DISCRETE  0x07D0
#define MB_MAXSIZE_WRITE_DISCRETE 0x07B0
#define MB_MAXSIZE_READ_REGISTER  0x007D
#define MB_MAXSIZE_WRITE_REGISTER 0x007B

//! \brief errors that can happen in modbus protocol stack and modbus uart
enum class mb_err {
	SUCCESS,
	ERR_BAUDRATE, 	//!< baudrate cannot be set, error is too big
	ERR_INIT,       //!< a sub class was not inited before use
	ERR_ARGS,		//!< invalid arguments (maybe NULL)
	ERR_EXEC,		//!< undefined error while executing
	ERR_BUSY,		//!< call me later
	ERR_DMA,		//!< error with dma handling
};

//! \brief function codes of modbus protocol
enum class mb_func_code {
	read_coils 				= 0x01,	//!< 
	read_discrete_inputs 	= 0x02,	//!< 
	read_holding_registers 	= 0x03,	//!< 
	read_input_registers 	= 0x04,	//!< 
	write_single_coil 		= 0x05,	//!< 
	write_single_register 	= 0x06,	//!< 
	read_exception_status 	= 0x07,	//!< 
	diagnostics 			= 0x08,	//!< 
	get_comm_event_counter 	= 0x0B,	//!< 
	get_comm_event_log 		= 0x0C,	//!< 
	write_multiple_coils 	= 0x0F,	//!< 
	write_multiple_registers = 0x10,//!< 
	report_server_id 		= 0x11,	//!< 
	read_file_record 		= 0x14,	//!< 
	write_file_record 		= 0x15,	//!< 
	mask_write_register 	= 0x16,	//!< 
	readwrite_multiple_registers = 0x17,	//!<
	read_fifo_queue 		= 0x18,	//!< 
	encapsulated_interface_transport = 0x2B	//!<
};

//! \brief sub function codes of modbus protocol diagnostics
enum class mb_sub_func_code {
	return_query_data 						= 0x00,
	restart_communications_option 			= 0x01,
	return_diagnostic_register 				= 0x02,
	change_ascii_input_delimiter 			= 0x03,
	force_listen_only_mode					= 0x04,
	clear_counters_and_diagnostics_register	= 0x0A,
	return_bus_message_count				= 0x0B,
	return_bus_communication_error_count 	= 0x0C,
	return_bus_exception_error_count 		= 0x0D,
	return_slave_message_count 				= 0x0E,
	return_slave_no_response_count 			= 0x0F,
	return_slave_nak_count 					= 0x10,
	return_slave_busy_count 				= 0x11,
	return_bus_character_overrun_count 		= 0x12,
	clear_overrun_counter_and_flag 			= 0x14,
};

//! \brief byte/word positions in a modbus frame
enum class mb_frame_position {
	address 		= 0,	//!< same in tx/rx
	function_code 	= 1,	//!< same in tx/rx
	status			= 1,	//!< status from get event counter (2 bytes)
	sub_function	= 1,	//!< sub function of diagnostics (2 bytes)
	startaddress 	= 2, 	//!< 16bit startaddress (sometimes used in tx)
	bytecount 		= 2,	//!< 8bit bytecount for tx
	data_field		= 3,	//!< event counter from get event counter or datafield diagnostics(2 bytes)
	quantity 		= 4,	//!< rxed quantity to handle
	value 			= 4,	//!< rxed output/register value
	bytecount_rx 	= 6, 	//!< bytecount field in rx frame
};

//! \brief types of callback that can be registered
enum class mb_callback_type {
	coil,			//!< callback for read/set coils
	input,			//!< callback for read discrete inputs
	register_input,	//!< callback for read input register
	register_hold,	//!< callback for read/write hold register
};

//! \brief modbus excption codes
enum class mb_exc {
	none 				 = 0, //!< 
	illegal_function 	 = 1, //!< 
	illegal_data_access  = 2, //!< 
	illegal_data_value 	 = 3, //!< 
	slave_device_failure = 4, //!< 
	acknowledge 		 = 5, //!< 
	slave_device_busy 	 = 6, //!< 
	negative_acknowledge = 7, //!< 
	memory_parity_error  = 8, //!< 
};

//! \brief possible paritys in modbus
//! \note Even parity is required - default parity mode must be even parity
//!       it is recommended to support also No parity mode - the use of no parity requires 2 stop bits
enum class mb_parmode : uint8_t {
	EVENP_1S   = 1, 
	defaulting = 1,
	ODDP_1S    = 2, 
	NOP_2S     = 3,
	pmax       = 4,
};

//! \brief possible baudrates of modbus
//! \note 9600 bps and 19.2 Kbps are required and 19.2 is the required default
//!       Every implemented baud rate must be respected better than 1% in transmission situation
//!       and must accept an error of 2% in reception situation.
enum class mb_baudrate : uint8_t {
	b2400 = 0,
	b4800 = 1,
	b9600 = 2,
	b14400 = 3,
	b19200 = 4,
	defaulting = 4,
	b28800 = 5,
	b38400 = 6,
	b57600 = 7,
	b76800 = 8,
	b115200 = 9,
	b230400 = 10,
	b250000 = 11,
	b460800 = 12,
	b500000 = 13,
	b921600 = 14,
	b1000000 = 15,
	bmax = 16,
};

struct mb_setting {
	uint8_t     address;
	mb_baudrate baudrate;
	mb_parmode  parity;
};

// baudrate/100
static const uint16_t baudlut[] = {
	24,
	48,
	96,
	144,
	192,
	288,
	384,
	576,
	768,
	1152,
	2304,
	2500,
	4608,
	5000,
	9216,
	10000
};

#endif//_MB_DEFS_H_