/* 
 * This file is part of the Modbus IO distribution.
 * Copyright (c) 2020 Martin Wende.
 * 
 * This program is free software: you can redistribute it and/or modify  
 * it under the terms of the GNU General Public License as published by  
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "modbus.hpp"

#include <cstddef>

modbus::modbus() :
	#if MBSUPP_READ_COILS
	m_callback_coils(NULL),
	#endif
	#if MBSUPP_READ_DISCIN
	m_callback_discin(NULL),
	#endif
	#if MBSUPP_READ_HOLDING
	m_callback_holding(NULL),
	#endif
	#if MBSUPP_READ_INPUT
	m_callback_input(NULL),
	#endif
	m_uart_inst(NULL),
	m_address(MB_ADDRESS_DEFAULT),
	m_framelen(0),
	m_state(mb_state::idle),
	m_reinit_after_send(false),
	m_reinit_uart_after_send(false),
	m_busy_condition(false),
	m_listen_only_mode(false),
	m_bus_message_count(0),
	m_bus_communications_error_count(0),
	m_bus_exception_error_count(0),
	m_bus_slave_message_count(0),
	m_bus_slave_no_response_count(0),
	m_bus_slave_nak_count(0),
	m_bus_slave_busy_count(0),
	m_bus_character_overrun_count(0),
	m_comm_event_counter(0) {
}

mb_err modbus::init(i_mb_uart *uart_inst, mb_baudrate baudrate, mb_parmode parity) {
	
	m_uart_inst = uart_inst;
	m_baudrate  = baudrate;
	m_parity    = parity;
	return init(baudrate, parity, true);
}

mb_err modbus::init(mb_baudrate baudrate, mb_parmode parity, bool now) {
	
	m_state    = mb_state::idle;
	mb_err err = mb_err::SUCCESS;
	
	m_baudrate = baudrate;
	m_parity   = parity;
	
	if (false == now)
	{
		m_reinit_uart_after_send = true;
		return err;
	}
	
	err = m_uart_inst->init(baudrate, parity);
	if (mb_err::SUCCESS != err){
		return err;
	}
	err = m_uart_inst->register_callback_rx(callback_rx, this);
	if (mb_err::SUCCESS != err){
		return err;
	}
	err = m_uart_inst->register_callback_tx(callback_tx, this);
	return err;
}

void modbus::reinit() {
	reset_counter();
}

mb_err modbus::set_address(uint8_t address) {
	
	mb_err err;
	err = check_address(address);
	if (err != mb_err::SUCCESS)
	{
		return err;
	}
	
	m_address = address;
	return err;
}

mb_err modbus::check_address(uint8_t address)
{
	if ((0 == address) || (address > MB_ADDRESS_DEFAULT)) 
	{
		return mb_err::ERR_ARGS;
	}
	return mb_err::SUCCESS;
}

mb_err modbus::check_baud(mb_baudrate baudrate) {
	return m_uart_inst->check_baud(baudrate);
}

mb_err modbus::check_baud(mb_baudrate* baudrate, uint8_t baud_idx)
{
	uint8_t baud_idx_max = (uint8_t)mb_baudrate::bmax;
	
	if ((NULL == baudrate) || (baud_idx >= baud_idx_max))
	{
		return mb_err::ERR_ARGS;
	}
	
	mb_err err;
	mb_baudrate baudrate_testing = (mb_baudrate)baud_idx;
	err = check_baud(baudrate_testing);
	if (mb_err::SUCCESS == err)
	{
		*baudrate = baudrate_testing;
	}
	
	return err;
}

mb_err modbus::check_parity(mb_parmode* parity, uint8_t parity_idx)
{
	uint8_t parity_idx_max = (uint8_t)mb_parmode::pmax;
	
	if ((NULL == parity) || (parity_idx >= parity_idx_max))
	{
		return mb_err::ERR_ARGS;
	}
	
	*parity = (mb_parmode)parity_idx;
	return mb_err::SUCCESS;
}

mb_err modbus::register_callback(mb_callback_type type, i_mb_callback *callback) {

	switch(type) {
		#if MBSUPP_READ_COILS
		case mb_callback_type::coil:
			m_callback_coils = callback;
			break;
		#endif
		#if MBSUPP_READ_DISCIN
		case mb_callback_type::input:
			m_callback_discin = callback;
			break;
		#endif
		#if MBSUPP_READ_HOLDING
		case mb_callback_type::register_input:
			m_callback_input = callback;
			break;
		#endif
		#if MBSUPP_READ_INPUT
		case mb_callback_type::register_hold:
			m_callback_holding = callback;
			break;
		#endif
	}
	// unknown callback
	return mb_err::ERR_EXEC;
}

mb_err modbus::process() {
	
	mb_err retval;
	
	switch (m_state) {
		
		// quit idle and receive frame
		case mb_state::idle:
		// rearming receiving
		case mb_state::rearm:
		// send done, receive new frame
		case mb_state::send:
			if (m_reinit_after_send) {
				m_reinit_after_send = false;
				reinit();
			}
			if (m_reinit_uart_after_send) {
				m_reinit_uart_after_send = false;
				init(m_baudrate, m_parity, true);
			}
			retval = m_uart_inst->receive(m_frame, sizeof(m_frame));
			if (mb_err::SUCCESS == retval){
				m_state = mb_state::recwait;
			} else {
				m_state = mb_state::err;
			}
			return retval;
		
		// wait a little bit more
		case mb_state::recwait:
			return mb_err::SUCCESS;
		
		// we have something
		case mb_state::reced:
			m_state = mb_state::handling;
			return decode();
		
		// wait a little bit more
		case mb_state::handling:
		case mb_state::sending:
			return mb_err::SUCCESS;
		
		case mb_state::err:
			return mb_err::ERR_EXEC;
	}
	
	// we shouldnt be here
	return mb_err::ERR_EXEC;
}

void modbus::can_sleep(bool *can_sleep)
{
	if (NULL == can_sleep)
	{
		return;
	}
	
	(*can_sleep) &= (mb_state::idle == m_state) || (mb_state::recwait == m_state);
}

void modbus::callback_rx(uint8_t* buf, uint32_t len, uint32_t err, void* arg) {
	(void)buf;
	
	// spurious call
	if (NULL == arg){
		return;
	}
	
	modbus *me = (modbus*)arg;
	
	// errs?
	if (err){
		me->m_state = mb_state::idle;
		me->m_bus_communications_error_count++;
		me->m_eventbytes.add_recv_event_comm_err();
		if (err & ((uint32_t)mb_uart_err::ERR_OVERRUN)) {
			me->m_eventbytes.add_recv_event_char_overrun();
			me->m_bus_character_overrun_count++;
		}
		me->m_eventbytes.add_recv_event_flush();
		return;
	}
	
	// we received without errors, set len and state
	if (mb_state::recwait == me->m_state) {
		me->m_framelen = len;
		me->m_state = mb_state::reced;
	}
}
	
void modbus::callback_tx(uint8_t* buf, uint32_t len, uint32_t err, void* arg) {
	(void)len;
	(void)buf;
	
	// spurious call
	if (NULL == arg){
		return;
	}
	
	modbus *me = (modbus*)arg;
	
	// errs?
	if (err){
		me->m_state = mb_state::idle;
		// TODO: errorcounter
		return;
	}
	
	// we had send
	if (mb_state::sending == me->m_state) {
		me->m_state = mb_state::send;
	}
}

mb_err modbus::decode() {
	
	// errors while receiving? -> receive next and dont answer
	if (m_uart_err != mb_uart_err::SUCCESS){
		m_state = mb_state::rearm;
		return mb_err::SUCCESS;
	}
	
	// has the frame the minimum size?
	if (m_framelen < MB_FRAME_SIZEMIN){
		m_bus_communications_error_count++;
		m_state = mb_state::rearm;
		return mb_err::SUCCESS;
	}
	
	// is the CRC ok?
	uint16_t crc = get_rx_crc16();
	if (false == crc_checkseq(m_frame, m_framelen - sizeof(uint16_t), crc)){
		m_bus_communications_error_count++;
		m_state = mb_state::rearm;
		return mb_err::SUCCESS;
	}
	
	m_bus_message_count++; // Messages with bad CRC are not taken into account
	
	// is this our address?
	uint8_t received_address = get_rx_address();
	if ((received_address != 0) && (received_address != m_address)){
		m_state = mb_state::rearm;
		return mb_err::SUCCESS;
	}
	if (0 == received_address) {
		m_eventbytes.add_recv_event_broadcast();
		m_bus_slave_no_response_count++;
	}
	// monitors but doesnt do anything more
	if (m_listen_only_mode) {
		m_eventbytes.add_recv_event_listen_only();
		m_eventbytes.add_recv_event_flush();
		m_state = mb_state::rearm;
		return mb_err::SUCCESS;
	}
	m_eventbytes.add_recv_event_flush();
	m_bus_slave_message_count++;
	
	mb_func_code function_code = get_rx_functioncode();
	
	switch (function_code) {
		
		#if MBSUPP_READ_COILS
		case mb_func_code::read_coils:
			return decode_read_bits(m_callback_coils, function_code);
		#endif
		
		#if MBSUPP_READ_DISCIN
		case mb_func_code::read_discrete_inputs:
			return decode_read_bits(m_callback_discin, function_code);
		#endif
		
		#if MBSUPP_READ_HOLDING
		case mb_func_code::read_holding_registers:
			return decode_read_registers(m_callback_holding, function_code);
		#endif
		
		#if MBSUPP_READ_INPUT
		case mb_func_code::read_input_registers:
			return decode_read_registers(m_callback_input, function_code);
		#endif
		
		case mb_func_code::diagnostics:
			return handle_diagnostics();
		
		case mb_func_code::get_comm_event_counter:
			return handle_event_counter();
		
		case mb_func_code::get_comm_event_log:
			return handle_event_log();
		
		#if MBSUPP_READ_COILS
		case mb_func_code::write_single_coil:
			return decode_write_single_coil();
		
		case mb_func_code::write_multiple_coils:
			return decode_write_multiple_coils();
		#endif
		
		#if MBSUPP_READ_HOLDING
		case mb_func_code::write_single_register:    // holding of course
			return decode_write_single_register();
		
		case mb_func_code::write_multiple_registers: // holding of course
			return decode_write_multiple_register();
		#endif
		
		default:
			return send_exception(mb_exc::illegal_function);
	}
	
	// we shouldnt be here
	return send_exception(mb_exc::slave_device_failure);
}

// TODO: umbauen auf neue set_tx funktionen
mb_err modbus::send_exception(mb_exc code){
	
	// It comprises also the error detected in broadcast messages even if an
	// exception message is not returned in this case.
	m_bus_exception_error_count++;

	// dont send this was a broadcast
	uint8_t address = get_rx_address();
	if (0 == address){
		m_state = mb_state::rearm;
		return mb_err::SUCCESS;
	}
	
	m_eventbytes.add_sent_event(code, m_listen_only_mode);
	
	if (mb_exc::negative_acknowledge == code) {
		m_bus_slave_nak_count++;
	}
	if (mb_exc::slave_device_busy == code) {
		m_bus_slave_busy_count++;
	}
	
	m_frame[1] |= 0x80;
	m_frame[2]  = (uint8_t)code;
	
	uint16_t crc = crc_genseq((uint8_t*)m_frame, MB_FRAME_SIZEEXC - sizeof(uint16_t));
	m_frame[3] = crc;
	m_frame[4] = crc>>8;
	
	m_framelen = MB_FRAME_SIZEEXC;
	
	return sendout();
}