/* 
 * This file is part of the Modbus IO distribution.
 * Copyright (c) 2020 Martin Wende.
 * 
 * This program is free software: you can redistribute it and/or modify  
 * it under the terms of the GNU General Public License as published by  
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef MODBUS_HPP
#define MODBUS_HPP

#include <cstdint>

#include "mb_defs.hpp"
#include "modbus_eventbytes.hpp"
#include "i_mb_uart.hpp"
#include "i_mb_callback.hpp" 

#include "i_sleepable.hpp"

// set the callbacks this device supports (write is included)
#define MBSUPP_READ_COILS 	1
#define MBSUPP_READ_DISCIN 	1
#define MBSUPP_READ_HOLDING	1
#define MBSUPP_READ_INPUT	1

class modbus : public i_sleepable {
public:
	
	//! \brief static constructor
	modbus();
	
	//! \brief init with uart instance
	//! \param[in] uart_inst uart instance
	//! \param[in] baudrate baudrate to set
	//! \param[in] parity parity to set
	//! \return @ref mb_err
	mb_err init(i_mb_uart *uart_inst, mb_baudrate baudrate, mb_parmode parity);
	
	//! \brief init with uart mode
	//! \param[in] baudrate baudrate to set
	//! \param[in] parity parity to set
	//! \return @ref mb_err
	mb_err init(mb_baudrate baudrate, mb_parmode parity, bool now);
	
	void reinit(void);
	
	//! \brief set address of this device
	//! \param[in] address address of this device
	//! \return @ref mb_err
	mb_err set_address(uint8_t address);
	
	//! \brief checks for legal address
	//! \param[in] address address of this device
	//! \return @ref mb_err
	mb_err check_address(uint8_t address);
	
	//! \brief check if baudrate is possible before set
	//! \param[in] baudrate baudrate to set
	//! \return @ref mb_err
	mb_err check_baud(mb_baudrate baudrate);
	
	//! \brief check if baudrate is possible before set, additional gets enum index and checks bounds
	//! \param[out] baudrate baudrate to set
	//! \param[in] baud_idx idx of enum from remote
	//! \return @ref mb_err
	mb_err check_baud(mb_baudrate* baudrate, uint8_t baud_idx);
	
	//! \brief check if parity is possible before set, additional gets enum index and checks bounds
	//! \param[out] parity parity to set
	//! \param[in] parity_idx idx of enum from remote
	//! \return @ref mb_err
	mb_err check_parity(mb_parmode* parity, uint8_t parity_idx);
	
	//! \brief register a callback for handling incoming frames
	//! \param[in] type type of this callback
	//! \param[in] callback callback to call for the given function
	//! \return @ref mb_err
	mb_err register_callback(mb_callback_type type, i_mb_callback *callback);
	
	//! \brief call this in main/task loop to handle modbus frames
	//! \return @ref mb_err
	mb_err process();
	
	virtual void can_sleep(bool *can_sleep);
	
private:

	// handle comm department------------------------------------------------------
	
	//! \brief mb uart callback
	//! \param[in] buf buffer that was received/send
	//! \param[in] len len that was received/send
	//! \param[in] err errflags that happened while sending/receiving
	//! \param[in] arg userarg given
	static void callback_rx(uint8_t* buf, uint32_t len, uint32_t err, void* arg);
	
	//! \brief mb uart callback
	//! \param[in] buf buffer that was received/send
	//! \param[in] len len that was received/send
	//! \param[in] err errflags that happened while sending/receiving
	//! \param[in] arg userarg given
	static void callback_tx(uint8_t* buf, uint32_t len, uint32_t err, void* arg);

	// decode department ----------------------------------------------------------

	//! \brief decodes a received frame and sends an answer
	//! \return @ref mb_err
	mb_err decode();
	
	#if (MBSUPP_READ_COILS || MBSUPP_READ_DISCIN)
	//! \brief decodes commands that are reading multiple bits
	//! \param[in] callback callback to use for reading bits
	//! \param[in] func function code to send back
	//! \return @ref mb_err
	mb_err decode_read_bits(i_mb_callback *callback, mb_func_code func);
	#endif
	
	#if (MBSUPP_READ_HOLDING || MBSUPP_READ_INPUT)
	//! \brief decodes commands that are reading multiple registers
	//! \param[in] callback callback to use for reading bits
	//! \param[in] func function code to send back
	//! \return @ref mb_err
	mb_err decode_read_registers(i_mb_callback *callback, mb_func_code func);
	#endif
	
	//! \brief handles the diagnostics command
	//! \return @ref mb_err
	mb_err handle_diagnostics();
	
	//! \brief handles the event counter command
	//! \return @ref mb_err
	mb_err handle_event_counter();
	
	//! \brief handles the event log command
	//! \return @ref mb_err
	mb_err handle_event_log();
	
	#if MBSUPP_READ_COILS
	//! \return @ref mb_err
	mb_err decode_write_single_coil();
	
	//! \return @ref mb_err
	mb_err decode_write_multiple_coils();
	#endif
	
	#if MBSUPP_READ_HOLDING
	//! \return @ref mb_err
	mb_err decode_write_single_register();
	
	//! \return @ref mb_err
	mb_err decode_write_multiple_register();
	#endif
	
	//! \brief sends an exception
	//! \param[in] code exceptioncode @ref mb_exc
	//! \return @ref mb_err
	mb_err send_exception(mb_exc code);
	
	// get rx fields department ---------------------------------------------------
	
	//! \brief gets the rx address out of the rx frame
	//! \return address
	uint8_t get_rx_address();
	
	//! \brief gets the function code out of the rx frame
	//! \return @ref mb_func_code
	mb_func_code get_rx_functioncode();
	
	//! \brief gets the sub function code out of the rx frame for diagnostics
	//! \return @ref mb_sub_func_code
	mb_sub_func_code get_rx_sub_function();
	
	//! \brief gets the data field of diagnostics
	//! \return data field
	uint16_t get_rx_data_field();
	
	//! \brief gets the start address of the virtual memory
	//! \return startaddress
	uint16_t get_rx_startaddress();
	
	//! \brief gets the quantity to read/write
	//! \return quantity
	uint16_t get_rx_quantity();
	
	//! \brief gets the value to write
	//! \return quantity
	uint16_t get_rx_value();
	
	//! \brief gets the bytecount to follow
	//! \return bytecount
	uint8_t get_rx_bytecount();
	
	//! \brief gets the startaddress of the (danyamic) payload
	//! \param[in] func function code to send back
	//! \return startaddress of the (danyamic) payload
	uint16_t get_tx_payloadstart(mb_func_code func);
	
	//! \brief gets the startaddress of the (danyamic) payload
	//! \param[in] func function code to send back
	//! \return startaddress of the (danyamic) payload
	uint16_t get_rx_payloadstart(mb_func_code func);
	
	// set tx fields department ---------------------------------------------------
	
	//! \brief sets the address field in tx frame
	//! \param[in] address address to set
	void set_tx_address(uint8_t address);
	
	//! \brief sets the function field in tx frame
	//! \param[in] code code to set
	void set_tx_function(mb_func_code code);
	
	//! \brief gets the sub function code out of the rx frame for diagnostics
	//! \param[in] code code to set
	void set_tx_sub_function(mb_sub_func_code code);
	
	//! \brief sets the startaddressfield in tx frame
	//! \param[in] startaddress startaddress to set
	void set_tx_startaddress(uint16_t startaddress);
	
	//! \brief sets the quantity field in tx frame
	//! \param[in] quantity quantity to set
	void set_tx_quantity(uint16_t quantity);
	
	//! \brief sets the value field in tx frame
	//! \param[in] value value to set
	void set_tx_value(uint16_t value);
	
	//! \brief sets the status field in tx frame
	//! \param[in] value value to set
	void set_tx_status(uint16_t status);
	
	//! \brief sets the data field in tx frame
	//! \param[in] value value to set
	void set_tx_data_field(uint16_t data);
	
	//! \brief sets the byte count field in tx frame
	//! \param[in] num_bytes num of bytes to set this field to
	void set_tx_byte_count(uint8_t num_bytes);
	
	//! \brief calculates and sets the crc of this frame, also sets final m_framelen
	//! \param[in] offset offset to put crc to
	void set_tx_crc(uint16_t offset);
	
	//! \brief tries to send and sets state + err 
	//! \return @ref mb_err
	mb_err sendout();
	
	// crc department -------------------------------------------------------------
	
	//! \brief sets the CRC init value
	void crc_start();
	
	//! \brief calculates the crc for one byte
	//! \param[in] chara byte to calculate over
	void crc_calc_char(uint8_t chara);
	
	//! \brief checks the calculated crc with the given crc
	//! \param[in] crc crc value to check against
	//! \return true if both are the same, else false
	bool crc_check(uint16_t crc);
	
	//! \brief calculates the crc over the given sequence and checks against the given crc
	//! \param[in] buf buffer to calculate over
	//! \param[in] len how many bytes to calculate
	//! \param[in] crc crc value to check against
	//! \return true if both are the same, else false
	bool crc_checkseq(uint8_t *buf, uint8_t len, uint16_t crc);
	
	//! \brief calculates the crc over the given sequence
	//! \param[in] buf buffer to calculate over
	//! \param[in] len how many bytes to calculate
	//! \return the CRC16
	uint16_t crc_genseq(uint8_t *buf, uint8_t len);
	
	//! \brief gets the cr16 from the incoming frame
	//! \return corresponding uint16_t CCR16
	uint16_t get_rx_crc16();
	
	// pack/unpack 16bits from/to buffer department -------------------------------
	
	//! \brief combines two elements from the buffer to a uint16_t
	//! \param[in] buf pointer to buffer
	//! \param[in] offset offset in buffer to start combining
	//! \return corresponding uint16_t
	uint16_t buf_to_u16(uint8_t *buf, uint16_t offset);
	
	//! \brief combines two elements from the buffer to a uint16_t
	//! \param[out] buf pointer to buffer
	//! \param[in] offset offset in buffer to start combining
	//! \param[in] val value to set
	void u16_to_buf(uint8_t *buf, uint16_t offset, uint16_t val);
	
	void reset_counter();
	
	// class variables department -------------------------------------------------
	
	//! \brief states of modbus client
	enum class mb_state {
		idle,		//!< idle
		rearm,      //!< rearm RX without using idle (this would flag a sleepable)
		recwait,	//!< waiting for a received frame
		reced,		//!< received a frame, wait for next process() call
		handling,	//!< received frame is handled
		sending,	//!< answer is send
		send,		//!< answer send is done, wait for next process() call
		err,		//!< error state, nothing will be done anymore
	};
	
	//! \brief the state of the modbus client now present
	mb_state m_state;
	
	//! \brief uart instance to use
	i_mb_uart *m_uart_inst;
	
	//! \brief the address of this modbus client
	uint8_t m_address;
	
	//! \brief uart modes
	mb_baudrate m_baudrate;
	mb_parmode  m_parity;
	
	//! \brief the modbus frame itself
	uint8_t m_frame[MB_FRAME_SIZEMAX];
	//! \brief the length of the modbus frame
	uint16_t m_framelen;
	//! \brief last receive uart err
	mb_uart_err m_uart_err;
	
	//! \brief the crc value after x crced bytes
	uint16_t m_crc_val;
	
	// diangnostics department ----------------------------------------------------
	
	//! \brief event bytes subclass
	modbus_eventbytes m_eventbytes;
	
	//! \brief reset counter after send
	bool m_reinit_after_send;
	
	//! \brief use new uart settings after send
	bool m_reinit_uart_after_send;
	
	//! \brief true if if a previouslyissued program command is still being processed by the remote device (a busy condition exists).
	bool m_busy_condition;
	
	//! \brief only listen to modbus, but dont answer and dont execute messages
	bool m_listen_only_mode;
	
	//! \brief the quantity of messages that the remote device has detected on the communications system since its last restart, clear counters operation, or powerup.
	uint16_t m_bus_message_count;
	
	//! \brief the quantity of CRC errors encountered by the remote device since its last restart, clear counters operation, or powerup.
	//! \note In case of an error detected on the character level, (overrun, parity error), or in case of a
	//! \note message length < 3 bytes, the receiving device is not able to calculate the CRC. In such cases, this counter is also incremented.
	uint16_t m_bus_communications_error_count;
	
	//! \brief the quantity of MODBUS exception responses returned by the remote device since its last restart, clear counters operation, or powerup.
	//! \note It comprises also the error detected in broadcast messages even if an exception message is not returned in this case.
	uint16_t m_bus_exception_error_count;
	
	//! \brief the quantity of messages addressed to the remote device, or broadcast, that the remote device has processed since its last restart, clear counters operation, or powerup.
	uint16_t m_bus_slave_message_count;
	
	//! \brief the quantity of messages addressed to the remote device for which it has returned no response (neither a normal response nor an exception response), since its last restart, clear counters operation, or powerup.
	//! \note this counter counts the number of broadcast messages it has received.
	uint16_t m_bus_slave_no_response_count;
	
	//! \brief the quantity of messages addressed to the remote device for which it returned a Negative Acknowledge (NAK) exception response, since its last restart, clear counters operation, or powerup.
	uint16_t m_bus_slave_nak_count;
	
	//! \brief the quantity of messages addressed to the remote device for which it returned a Slave Device Busy exception response, since its last restart, clear counters operation, or powerup.
	uint16_t m_bus_slave_busy_count;
	
	//! \brief the quantity of messages addressed to the remote device that it could not handle due to a character overrun condition, since its last restart, clear counters operation, or powerup.
	uint16_t m_bus_character_overrun_count;
	
	//! \brief The devices event counter is incremented once for each successful message completion. It is not incremented for exception responses, poll commands, or fetch event counter commands.
	uint16_t m_comm_event_counter;
	
	uint16_t m_diagnostic_register; // TODO: was steht da drinne?
	
	// callbacks department -------------------------------------------------------
	
	#if MBSUPP_READ_COILS
	i_mb_callback *m_callback_coils;
	#endif
	
	#if MBSUPP_READ_DISCIN
	i_mb_callback *m_callback_discin;
	#endif
	
	#if MBSUPP_READ_HOLDING
	i_mb_callback *m_callback_holding;
	#endif
	
	#if MBSUPP_READ_INPUT
	i_mb_callback *m_callback_input;
	#endif
	
};

#endif//MODBUS_HPP