/* 
 * This file is part of the Modbus IO distribution.
 * Copyright (c) 2020 Martin Wende.
 * 
 * This program is free software: you can redistribute it and/or modify  
 * it under the terms of the GNU General Public License as published by  
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "modbus.hpp"

#include <cstddef>

#if (MBSUPP_READ_COILS || MBSUPP_READ_DISCIN)
mb_err modbus::decode_read_bits(i_mb_callback *callback, mb_func_code func) {
	
	uint16_t startaddress = get_rx_startaddress();
	uint16_t quantity = get_rx_quantity();
	
	// if no callback is registered, then this function is illegal
	if (NULL == callback) {
		return send_exception(mb_exc::illegal_function);
	}
	
	// is he asking for too much?
	if ((0 == quantity) || (quantity > MB_MAXSIZE_READ_DISCRETE)) {
		return send_exception(mb_exc::illegal_data_value);
	}
	
	// address out of bounds?
	if ((startaddress + quantity - 1) > callback->m_max_address) {
		return send_exception(mb_exc::illegal_data_access);
	}
	
	// round up to upper limit (1 coil needs still 1 byte)
	uint8_t num_bytes = (quantity + 7)/8;
	
	uint8_t set_byte = 0, bit_idx;
	bool last_byte_was_written = false;
	mb_err err;
	uint16_t bit, set;
	uint16_t frame_idx = get_tx_payloadstart(func);
	
	for (bit = 0; bit < quantity; bit++) {
		
		bit_idx = bit%8;
		
		// new start
		if (0 == bit_idx){
			set_byte = 0;
			last_byte_was_written = false;
		}
		
		set = 0; // myabe a callback dosent even write to the pointer if not set
		err = callback->call_read(&set, 1 + startaddress++);
		if (mb_err::SUCCESS != err) {
			return send_exception(mb_exc::slave_device_failure);
		}
		
		if (set) {
			set_byte |= 1 << bit_idx;
		}
		
		// end
		if (7 == bit_idx){
			m_frame[frame_idx++] = set_byte;
			last_byte_was_written = true;
		}
	}
	
	// there not always a full 8 bit cycle
	if (!last_byte_was_written) {
		m_frame[frame_idx++] = set_byte;
	}
	
	m_comm_event_counter++;
	
	// build frame
	set_tx_address(m_address);
	set_tx_function(func);
	set_tx_byte_count(num_bytes);
	set_tx_crc(frame_idx);
	return sendout();
}
#endif

#if MBSUPP_READ_COILS
mb_err modbus::decode_write_single_coil() {
	
	uint16_t startaddress = get_rx_startaddress();
	uint16_t value = get_rx_value();
	uint16_t set = (0xFF00 == value);
	
	// if no callback is registered, then this function is illegal
	if (NULL == m_callback_coils) {
		return send_exception(mb_exc::illegal_function);
	}
	
	// address out of bounds?
	if ((startaddress) > m_callback_coils->m_max_address) {
		return send_exception(mb_exc::illegal_data_access);
	}
	
	mb_err err;
	err = m_callback_coils->call_write(set, 1 + startaddress);
	if (mb_err::SUCCESS != err) {
		return send_exception(mb_exc::slave_device_failure);
	}
	
	m_comm_event_counter++;
	
	// build frame
	set_tx_address(m_address);
	set_tx_function(mb_func_code::write_single_coil);
	set_tx_startaddress(startaddress);
	set_tx_value(value);
	uint16_t frame_idx = get_tx_payloadstart(mb_func_code::write_single_coil);
	set_tx_crc(frame_idx);
	return sendout();
}

mb_err modbus::decode_write_multiple_coils() {

	uint16_t startaddress = get_rx_startaddress();
	uint16_t quantity = get_rx_quantity();
	uint16_t pay_offset = get_rx_payloadstart(mb_func_code::write_multiple_coils);
	uint8_t bytecount = get_rx_bytecount();
	
	// round up to upper limit (1 coil needs still 1 byte)
	uint8_t num_bytes = (quantity + 7)/8;
	
	// if no callback is registered, then this function is illegal
	if (NULL == m_callback_coils) {
		return send_exception(mb_exc::illegal_function);
	}
	
	// is he asking for too much?
	if ((0 == quantity) || (quantity > MB_MAXSIZE_WRITE_DISCRETE)) {
		return send_exception(mb_exc::illegal_data_value);
	}
	
	// is bytecount fitting to quantity?
	if (bytecount != num_bytes) {
		return send_exception(mb_exc::illegal_data_value);
	}
	
	// address out of bounds?
	if ((startaddress + quantity - 1) > m_callback_coils->m_max_address) {
		return send_exception(mb_exc::illegal_data_access);
	}
	
	uint8_t set_byte = 0, bit_idx;
	uint16_t  set;
	mb_err err;
	uint16_t bit;
	
	for (bit = 0; bit < quantity; bit++) {
		
		bit_idx = bit%8;
		
		// new start
		if (0 == bit_idx){
			set_byte = m_frame[pay_offset++];
		}
		
		set = (set_byte & (1 << bit_idx));
		
		err = m_callback_coils->call_write(set, 1 + startaddress++);
		if (mb_err::SUCCESS != err) {
			return send_exception(mb_exc::slave_device_failure);
		}
	}
	
	m_comm_event_counter++;
	
	// build frame
	set_tx_address(m_address);
	set_tx_function(mb_func_code::write_multiple_coils);
	set_tx_startaddress(startaddress);
	set_tx_quantity(quantity);
	uint16_t frame_idx = get_tx_payloadstart(mb_func_code::write_multiple_coils);
	set_tx_crc(frame_idx);
	return sendout();
}
#endif