/* 
 * This file is part of the Modbus IO distribution.
 * Copyright (c) 2020 Martin Wende.
 * 
 * This program is free software: you can redistribute it and/or modify  
 * it under the terms of the GNU General Public License as published by  
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "modbus_eventbytes.hpp"

modbus_eventbytes::modbus_eventbytes() :
	m_recv_byte(0),
	m_fifo_wp(0),
	m_fifo_fill(0) {
}

void modbus_eventbytes::add_recv_event_comm_err() {
	m_recv_byte |= (uint8_t)recv_bits::communication_error;
}

void modbus_eventbytes::add_recv_event_char_overrun() {
	m_recv_byte |= (uint8_t)recv_bits::character_overrun;
}

void modbus_eventbytes::add_recv_event_listen_only() {
	m_recv_byte |= (uint8_t)recv_bits::listen_only_mode;
}

void modbus_eventbytes::add_recv_event_broadcast() {
	m_recv_byte |= (uint8_t)recv_bits::broadcast_received;
}

void modbus_eventbytes::add_recv_event_flush() {
	m_recv_byte |= (uint8_t)recv_bits::receive_event;
	set_fifo(m_recv_byte);
	m_recv_byte = 0x00;
}

void modbus_eventbytes::add_sent_event(mb_exc code, bool listen_only_mode) {
	uint8_t sent_byte = 0x00;
	
	switch (code) { 
		case mb_exc::illegal_function:
		case mb_exc::illegal_data_access:
		case mb_exc::illegal_data_value:
			sent_byte |= (uint8_t)sent_bits::read_exception;
			break;
		case mb_exc::slave_device_failure:
			sent_byte |= (uint8_t)sent_bits::slave_abort_exception;
			break;
		case mb_exc::acknowledge:
		case mb_exc::slave_device_busy:
			sent_byte |= (uint8_t)sent_bits::slave_busy_exception;
			break;
		case mb_exc::negative_acknowledge:
			sent_byte |= (uint8_t)sent_bits::slave_program_nak_exception;
			break;
		default:
			break;
	}
	
	if (listen_only_mode) {
		sent_byte |= (uint8_t)sent_bits::listen_only_mode;
	}
	sent_byte |= (uint8_t)sent_bits::sent_event;
	set_fifo(sent_byte);
}
	
void modbus_eventbytes::add_enter_listen_only() {
	set_fifo(m_entered_listen_only);
}
	
void modbus_eventbytes::add_inited_comm_restart() {
	set_fifo(m_inited_comm_restart);
}

uint8_t modbus_eventbytes::get_fifo(uint8_t *bytes) {
	
	uint8_t m_fifo_rp = m_fifo_wp - 1;
	
	uint32_t i;
	for (i = 0; i < m_fifo_fill; i++) {
		if (m_fifo_rp >= fifosize) {
			m_fifo_rp  = fifosize - 1;
		}
		bytes[i] = m_fifo[m_fifo_rp--];
	}
	
	return m_fifo_fill;
}

void modbus_eventbytes::clear() {
	m_recv_byte = 0;
	m_fifo_wp = 0;
	m_fifo_fill = 0;
}

void modbus_eventbytes::set_fifo(uint8_t byte) {
	
	m_fifo[m_fifo_wp++] = byte;
	
	if (m_fifo_wp >= fifosize) {
		m_fifo_wp  = 0;
	}
	
	m_fifo_fill++;
	if (m_fifo_fill > fifosize) {
		m_fifo_fill = fifosize;
	}
}