/* 
 * This file is part of the Modbus IO distribution.
 * Copyright (c) 2020 Martin Wende.
 * 
 * This program is free software: you can redistribute it and/or modify  
 * it under the terms of the GNU General Public License as published by  
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "modbus.hpp"

#define MB_CRC_POLY 0xA001

uint8_t modbus::get_rx_address() {
	uint8_t offset = (uint8_t)mb_frame_position::address;
	return m_frame[offset];
}
	
mb_func_code modbus::get_rx_functioncode() {
	uint8_t offset = (uint8_t)mb_frame_position::function_code;
	uint8_t funcode_uint = m_frame[offset];
	return (mb_func_code)funcode_uint;
}

mb_sub_func_code modbus::get_rx_sub_function() {
	uint8_t offset = (uint8_t)mb_frame_position::sub_function;
	uint8_t funcode_uint = m_frame[offset];
	return (mb_sub_func_code)funcode_uint;
}

uint16_t modbus::get_rx_data_field() {
	uint8_t offset = (uint8_t)mb_frame_position::data_field;
	return buf_to_u16(m_frame, offset);
}

uint16_t modbus::get_rx_startaddress() {
	uint16_t offset = (uint16_t)mb_frame_position::startaddress;
	return buf_to_u16(m_frame, offset);
}
	
uint16_t modbus::get_rx_quantity() {
	uint16_t offset = (uint16_t)mb_frame_position::quantity;
	return buf_to_u16(m_frame, offset);
}

uint16_t modbus::get_rx_value() {
	uint16_t offset = (uint16_t)mb_frame_position::value;
	return buf_to_u16(m_frame, offset);
}

uint8_t modbus::get_rx_bytecount() {
	uint16_t offset = (uint16_t)mb_frame_position::bytecount_rx;
	return m_frame[offset];
}

uint16_t modbus::get_tx_payloadstart(mb_func_code func) {
	
	switch (func) {
		
		case mb_func_code::read_coils:
			return 3;
		case mb_func_code::read_discrete_inputs:
			return 3;
		case mb_func_code::read_holding_registers:
			return 3;
		case mb_func_code::read_input_registers:
			return 3;
		case mb_func_code::write_single_coil:
			return 6; // no dynamic payload, this is the crc start
		case mb_func_code::write_single_register:
			return 6; // no dynamic payload, this is the crc start
		case mb_func_code::diagnostics:
			return 6; // no dynamic payload, this is the crc start
		case mb_func_code::get_comm_event_counter:
			return 6; // no dynamic payload, this is the crc start
		case mb_func_code::write_multiple_coils:
			return 6; // no dynamic payload, this is the crc start
		case mb_func_code::write_multiple_registers:
			return 6; // no dynamic payload, this is the crc start
		
		default:
			return 0;
	}
}

uint16_t modbus::get_rx_payloadstart(mb_func_code func) {
	
	switch (func) {
		
		case mb_func_code::read_coils:
			return 6; // no dynamic payload, this is the crc start
		case mb_func_code::read_discrete_inputs:
			return 6; // no dynamic payload, this is the crc start
		case mb_func_code::read_holding_registers:
			return 6; // no dynamic payload, this is the crc start
		case mb_func_code::read_input_registers:
			return 6; // no dynamic payload, this is the crc start
		case mb_func_code::write_single_coil:
			return 6; // no dynamic payload, this is the crc start
		case mb_func_code::write_single_register:
			return 6; // no dynamic payload, this is the crc start
		case mb_func_code::write_multiple_coils:
			return 7;
		case mb_func_code::write_multiple_registers:
			return 7;
		
		default:
			return 0;
	}
}

void modbus::set_tx_address(uint8_t address) {
	uint8_t offset = (uint8_t)mb_frame_position::address;
	m_frame[offset] = address;
}

void modbus::set_tx_function(mb_func_code code) {
	uint8_t offset = (uint8_t)mb_frame_position::function_code;
	m_frame[offset] = (uint8_t)code;
}

void modbus::set_tx_sub_function(mb_sub_func_code code) {
	uint8_t offset = (uint8_t)mb_frame_position::sub_function;
	uint16_t value = (uint16_t)code;
	u16_to_buf(m_frame, offset, value);
}

void modbus::set_tx_startaddress(uint16_t startaddress) {
	uint8_t offset = (uint8_t)mb_frame_position::startaddress;
	u16_to_buf(m_frame, offset, startaddress);
}

void modbus::set_tx_quantity(uint16_t quantity) {
	uint16_t offset = (uint16_t)mb_frame_position::quantity;
	u16_to_buf(m_frame, offset, quantity);
}

void modbus::set_tx_value(uint16_t value) {
	uint16_t offset = (uint16_t)mb_frame_position::value;
	u16_to_buf(m_frame, offset, value);
}

void modbus::set_tx_status(uint16_t status) {
	uint16_t offset = (uint16_t)mb_frame_position::status;
	u16_to_buf(m_frame, offset, status);
}
	
void modbus::set_tx_data_field(uint16_t data) {
	uint16_t offset = (uint16_t)mb_frame_position::data_field;
	u16_to_buf(m_frame, offset, data);
}

void modbus::set_tx_byte_count(uint8_t num_bytes) {
	uint8_t offset = (uint8_t)mb_frame_position::bytecount;
	m_frame[offset] = num_bytes;
}

void modbus::set_tx_crc(uint16_t offset) {
	uint16_t crc = crc_genseq(m_frame, offset);
	// u16_to_buf not used, crc has other endianess
	m_frame[offset + 0] = crc;
	m_frame[offset + 1] = crc >> 8;
	m_framelen = offset + sizeof(uint16_t);
}

mb_err modbus::sendout() {
	
	mb_err retval;
	
	// dont send this was a broadcast
	uint8_t address = get_rx_address();
	if (0 == address){
		m_state = mb_state::idle;
		return mb_err::SUCCESS;
	}
	
	retval = m_uart_inst->send(m_frame, m_framelen);
	if (mb_err::SUCCESS == retval){
		m_state = mb_state::sending;
	}else{
		m_state = mb_state::err;
	}
	return retval;
}

void modbus::crc_start() {

	// The CRC calculation is started by first pre-loading a 16bit register to all 1s
	m_crc_val = 0xFFFF;
}

void modbus::crc_calc_char(uint8_t chara) {

	uint8_t i;
	uint8_t lasb_was;
	
	// During generation of the CRC, each 8bit character is exclusive ORed with the register contents
	m_crc_val ^= chara;
	
	for (i = 0; i < 8; i++){
	
		// The LSB is extracted and examined.
		lasb_was = m_crc_val & 1;
		
		// Then the result is shifted in the direction of the least significant bit (LSB)
		// with a zero filled into the most significant bit (MSB) position
		m_crc_val >>= 1;
		m_crc_val  &= 0x7FFF;
		
		//If the LSB was a 1, the register is then exclusive ORed with a preset, fixed value. 
		//If the LSB was a 0, no exclusive OR takes place.
		if (lasb_was){
			m_crc_val ^= MB_CRC_POLY;
		}
	}
}

bool modbus::crc_check(uint16_t crc) {
	return m_crc_val == crc;
}

bool modbus::crc_checkseq(uint8_t *buf, uint8_t len, uint16_t crc) {
	
	crc_genseq(buf, len);
	return crc_check(crc);
}

uint16_t modbus::crc_genseq(uint8_t *buf, uint8_t len) {
	
	crc_start();
	
	while (len--){
		crc_calc_char(*buf++);
	}
	
	return m_crc_val;
}

uint16_t modbus::get_rx_crc16() {
	// CRC16 is stored low to high
	uint16_t retval = 0;
	uint16_t crc_start = m_framelen - sizeof(uint16_t);
	retval |= m_frame[crc_start + 0];
	retval |= m_frame[crc_start + 1] << 8;
	return retval;
}

uint16_t modbus::buf_to_u16(uint8_t *buf, uint16_t offset) {
	// uint16_t is stored high to low
	uint16_t retval = 0;
	retval |= buf[offset + 0] << 8;
	retval |= buf[offset + 1];
	return retval;
}

void modbus::u16_to_buf(uint8_t *buf, uint16_t offset, uint16_t val){
	// uint16_t is stored high to low
	buf[offset + 0] = val >> 8;
	buf[offset + 1] = val;
}