/* 
 * This file is part of the Modbus IO distribution.
 * Copyright (c) 2020 Martin Wende.
 * 
 * This program is free software: you can redistribute it and/or modify  
 * it under the terms of the GNU General Public License as published by  
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "modbus.hpp"

#include <cstddef>

#if (MBSUPP_READ_HOLDING || MBSUPP_READ_INPUT)
mb_err modbus::decode_read_registers(i_mb_callback *callback, mb_func_code func) {
	
	uint16_t startaddress = get_rx_startaddress();
	uint16_t quantity = get_rx_quantity();
	
	// if no callback is registered, then this function is illegal
	if (NULL == callback) {
		return send_exception(mb_exc::illegal_function);
	}
	
	// is he asking for too much?
	if ((0 == quantity) || (quantity > MB_MAXSIZE_READ_REGISTER)) {
		return send_exception(mb_exc::illegal_data_value);
	}
	
	// address out of bounds?
	if ((startaddress + quantity - 1) > callback->m_max_address) {
		return send_exception(mb_exc::illegal_data_access);
	}
	
	uint16_t frame_idx = get_tx_payloadstart(func);
	uint16_t num_bytes = 0;
	uint16_t value;
	mb_err err;
	
	while (quantity--) {
		err = callback->call_read(&value, 1 + startaddress++);
		if (mb_err::SUCCESS != err) {
			return send_exception(mb_exc::slave_device_failure);
		}
		
		u16_to_buf(m_frame, frame_idx, value);
		frame_idx += sizeof(uint16_t);
		num_bytes += sizeof(uint16_t);
	}
	
	m_comm_event_counter++;
	
	// build frame
	set_tx_address(m_address);
	set_tx_function(func);
	set_tx_byte_count(num_bytes);
	set_tx_crc(frame_idx);
	return sendout();
}
#endif

#if MBSUPP_READ_HOLDING
mb_err modbus::decode_write_single_register() {
	
	uint16_t startaddress = get_rx_startaddress();
	uint16_t value = get_rx_value();
	
	// if no callback is registered, then this function is illegal
	if (NULL == m_callback_coils) {
		return send_exception(mb_exc::illegal_function);
	}
	
	// address out of bounds?
	if ((startaddress) > m_callback_coils->m_max_address) {
		return send_exception(mb_exc::illegal_data_access);
	}
	
	mb_err err;
	err = m_callback_holding->call_write(value, 1 + startaddress);
	if (mb_err::SUCCESS != err) {
		return send_exception(mb_exc::slave_device_failure);
	}
	
	m_comm_event_counter++;
	
	// build frame
	set_tx_address(m_address);
	set_tx_function(mb_func_code::write_single_register);
	set_tx_startaddress(startaddress);
	set_tx_value(value);
	uint16_t frame_idx = get_tx_payloadstart(mb_func_code::write_single_register);
	set_tx_crc(frame_idx);
	return sendout();
}

mb_err modbus::decode_write_multiple_register() {

	uint16_t startaddress = get_rx_startaddress();
	uint16_t quantity = get_rx_quantity();
	uint16_t pay_offset = get_rx_payloadstart(mb_func_code::write_multiple_registers);
	uint8_t byte_count = get_rx_bytecount();
	
	// if no callback is registered, then this function is illegal
	if (NULL == m_callback_holding) {
		return send_exception(mb_exc::illegal_function);
	}
	
	// is he asking for too much?
	if ((0 == quantity) || (quantity > MB_MAXSIZE_WRITE_DISCRETE)) {
		return send_exception(mb_exc::illegal_data_value);
	}
	
	// is bytecount fitting to quantity?
	if (byte_count != (2*quantity)) {
		return send_exception(mb_exc::illegal_data_value);
	}
	
	// address out of bounds?
	if ((startaddress + quantity - 1) > m_callback_holding->m_max_address) {
		return send_exception(mb_exc::illegal_data_access);
	}
	
	uint8_t reg;
	uint16_t value;
	mb_err err;
	
	for (reg = 0; reg < quantity; reg++) {
		
		value = buf_to_u16(m_frame, pay_offset);
		pay_offset += sizeof(uint16_t);
		
		err = m_callback_holding->call_write(value, 1 + startaddress++);
		if (mb_err::SUCCESS != err) {
			return send_exception(mb_exc::slave_device_failure);
		}
	}
	
	m_comm_event_counter++;
	
	// build frame
	set_tx_address(m_address);
	set_tx_function(mb_func_code::write_multiple_registers);
	set_tx_startaddress(startaddress);
	set_tx_quantity(quantity);
	uint16_t frame_idx = get_tx_payloadstart(mb_func_code::write_multiple_registers);
	set_tx_crc(frame_idx);
	return sendout();
}
#endif