/* 
 * This file is part of the Modbus IO distribution.
 * Copyright (c) 2020 Martin Wende.
 * 
 * This program is free software: you can redistribute it and/or modify  
 * it under the terms of the GNU General Public License as published by  
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "modbus.hpp"

#include <cstddef>

mb_err modbus::handle_diagnostics() {
	
	mb_sub_func_code subfunc = get_rx_sub_function();
	
	uint16_t retdata = 0x0000;
	
	switch (subfunc) {
		
		case mb_sub_func_code::return_query_data:
			retdata = get_rx_data_field();
			break;
		
		case mb_sub_func_code::restart_communications_option:
			retdata = get_rx_data_field();
			m_reinit_after_send = true;
			reset_counter();
			if (0xFF00 == retdata) {
				m_eventbytes.clear();
			}
			break;
		
		case mb_sub_func_code::return_diagnostic_register:
			retdata = m_diagnostic_register;
			break;
		
		//case mb_sub_func_code::change_ascii_input_delimiter: no ascii mode yet
		
		case mb_sub_func_code::force_listen_only_mode:
			m_listen_only_mode = true;
			break;
		
		case mb_sub_func_code::clear_counters_and_diagnostics_register:
			retdata = get_rx_data_field();
			reset_counter();
			m_diagnostic_register = 0;
			break;
		
		case mb_sub_func_code::return_bus_message_count:
			retdata = m_bus_message_count;
			break;
		
		case mb_sub_func_code::return_bus_communication_error_count:
			retdata = m_bus_communications_error_count;
			break;
		
		case mb_sub_func_code::return_bus_exception_error_count:
			retdata = m_bus_exception_error_count;
			break;
		
		case mb_sub_func_code::return_slave_message_count:
			retdata = m_bus_slave_message_count;
			break;
		
		case mb_sub_func_code::return_slave_no_response_count:
			retdata = m_bus_slave_no_response_count;
			break;
		
		case mb_sub_func_code::return_slave_nak_count:
			retdata = m_bus_slave_nak_count;
			break;
		
		case mb_sub_func_code::return_slave_busy_count:
			retdata = m_bus_slave_busy_count;
			break;
		
		case mb_sub_func_code::return_bus_character_overrun_count:
			retdata = m_bus_character_overrun_count;
			break;
		
		case mb_sub_func_code::clear_overrun_counter_and_flag:
			retdata = get_rx_data_field();
			m_bus_character_overrun_count = 0;
			break;
		
		default:
			return send_exception(mb_exc::illegal_function);
	}
	
	// send response
	// build frame
	set_tx_address(m_address);
	set_tx_function(mb_func_code::diagnostics);
	set_tx_sub_function(subfunc);
	set_tx_data_field(retdata);
	uint16_t frame_idx = get_tx_payloadstart(mb_func_code::diagnostics);
	set_tx_crc(frame_idx);
	mb_err err = sendout();
	
	// reinit after response is send
	// immediate if listen only mode
	if (m_listen_only_mode && m_reinit_after_send) {
		m_listen_only_mode = false;
		reinit();
	}
	
	return err;
}

mb_err modbus::handle_event_counter() {

	uint16_t status = 0x0000;
	if (m_busy_condition) {
		status = 0xFFFF;
	}

	// build frame
	set_tx_address(m_address);
	set_tx_function(mb_func_code::get_comm_event_counter);
	set_tx_status(status);
	set_tx_data_field(m_comm_event_counter);
	uint16_t frame_idx = get_tx_payloadstart(mb_func_code::get_comm_event_counter);
	set_tx_crc(frame_idx);
	return sendout();
}

mb_err modbus::handle_event_log() {
	return mb_err::SUCCESS;
}

void modbus::reset_counter() {
	m_bus_message_count = 0;
	m_bus_communications_error_count = 0;
	m_bus_exception_error_count = 0;
	m_bus_slave_message_count = 0;
	m_bus_slave_no_response_count = 0;	
	m_bus_slave_nak_count = 0;
	m_bus_slave_busy_count = 0;
	m_bus_character_overrun_count = 0;
	m_comm_event_counter = 0;
}
