#include "lm75_sleep.hpp"

#define LM75_MEASTIME_MS (300+50)
#define LM75_ADDR 0b10011110

lm75_sleep::lm75_sleep() :
	m_state(lm75_state::NoInit),
	m_temp(0-32767)
{
}	
	
lm75_err lm75_sleep::init(i_i2c* i2c_inst, pio_set* pin_power_en, i_timer_poll* timer_inst)
{
	m_state = lm75_state::NoInit;
	if ((NULL == i2c_inst) || (NULL == pin_power_en) || (NULL == timer_inst))
	{
		return lm75_err::ERR_ARGS;
	}
	
	m_i2c_inst     = i2c_inst;
	m_pin_power_en = pin_power_en;
	m_timer_inst   = timer_inst;
	m_state        = lm75_state::Idle;
	
	return lm75_err::SUCCESS;
}

void lm75_sleep::deinit(void)
{
	if (lm75_state::NoInit == m_state)
	{
		return;
	}
	
	m_pin_power_en->disable();
	m_state = lm75_state::NoInit;
}
	
lm75_err lm75_sleep::read_device(void)
{
	if (lm75_state::NoInit == m_state)
	{
		return lm75_err::ERR_NOINIT;
	}
	
	if (lm75_state::Idle == m_state)
	{
		m_state = lm75_state::Startup;
	}
	
	return lm75_err::SUCCESS;
}
	
lm75_err lm75_sleep::read_progress(bool* doing)
{
	if (NULL == doing)
	{
		return lm75_err::ERR_ARGS;
	}
	
	if (lm75_state::NoInit == m_state)
	{
		*doing = false;
		return lm75_err::ERR_NOINIT;
	}
	
	// if were here and its in idle then its doing nothing
	*doing = true;
	if (lm75_state::Idle == m_state)
	{
		*doing = false;
	}
	
	return lm75_err::SUCCESS;
}
	
lm75_err lm75_sleep::read_temp(int16_t* temp)
{
	if (NULL == temp)
	{
		return lm75_err::ERR_ARGS;
	}
	
	if (lm75_state::NoInit == m_state)
	{
		*temp = 0-32767;
		return lm75_err::ERR_NOINIT;
	}
	
	*temp = m_temp;
	return lm75_err::SUCCESS;
}
	
lm75_err lm75_sleep::process(void)
{
	bool expired;
	
	switch (m_state)
	{
		case lm75_state::NoInit:
			break;
		
		case lm75_state::Idle:
			break;
		
		case lm75_state::Startup:
			m_pin_power_en->enable();
			m_timer_inst->start(&m_timer_idx, LM75_MEASTIME_MS);
			m_state = lm75_state::WaitPowOn;
			break;
		
		case lm75_state::WaitPowOn:
			m_timer_inst->check_expiry(&m_timer_idx, &expired);
			if (expired)
			{
				m_state = lm75_state::Readout;
			}
			break;
		
		case lm75_state::Readout:
			read_and_save();
			m_pin_power_en->disable();
			m_state = lm75_state::Idle;
			break;
	}
	
	return lm75_err::SUCCESS;
}

void lm75_sleep::can_sleep(bool *can_sleep)
{
	if (NULL == can_sleep)
	{
		return;
	}
	
	(*can_sleep) &= lm75_state::Idle == m_state;
}

void lm75_sleep::read_and_save(void)
{
	uint8_t command = 0; // select register 0 = temperature 
	uint8_t data[2];
	
	i2c_err err;
	err = m_i2c_inst->read(LM75_ADDR, &command, sizeof(command), data, sizeof(data), true);
	if (i2c_err::SUCCESS != err)
	{
		m_temp = 0-32767;
		return;
	}
	
	uint16_t regval = data[0] << 8 | data[1];
	regval >>= 7; // 7 unused LSB
	int16_t temp = regval;
	
	// signex
	if (regval & (1 << 8))
	{
		// MSB is set -> signex
		temp = 0xFE00 | regval;
	}
	
	// 2k with LSB = 0,5degC to 2k with LSB = 0,1degC
	m_temp = temp * 5;
}