#include "water_barrel.hpp"

#include "cmsis_includer.h"

#define ADC_CHAN_SENS 1
#define ADC_CHAN_VCC  0

#define VOLT_FULLRANGE 5000

#define SENS_WARMUP_MS (20+10)

#include "math.h"

static const uint8_t adcChan[2] = {ADC_CHAN_SENS, ADC_CHAN_VCC};

water_barrel::water_barrel() :
	m_pin_power_en(NULL),
	m_state(wb_states::NoInit),
	m_values_mv{0},
	m_values_adc{0},
	m_water_level_mm(0),
	m_water_volume_ml(0)
{
}

barrel_err water_barrel::init(pio_set* pin_power_en, const barrel_calibdata* config, i_timer_poll* timer_inst)
{

	m_state = wb_states::NoInit;
	
	if ((NULL == pin_power_en) || (NULL == config) || (NULL == timer_inst))
	{
		return barrel_err::ERR_ARGS;
	}
	
	m_pin_power_en = pin_power_en;
	m_config       = config; 
	m_timer_inst   = timer_inst;

	m_state = wb_states::Idle;
	
	return barrel_err::ERR_NOINIT;
}

void water_barrel::deinit(void)
{
	if (wb_states::NoInit == m_state)
	{
		return;
	}
	
	m_pin_power_en->disable();
	
	switch (m_state)
	{
		case wb_states::NoInit:
		case wb_states::Idle:
		case wb_states::InitAdc:
			m_state = wb_states::NoInit;
			break;
		case wb_states::WaitAdcKalib:
			while (0 != (ADC1->CR & ADC_CR_ADCAL)) {asm volatile("nop");}
			adc_deinit();
			m_state = wb_states::NoInit;
			break;
		case wb_states::WaitAdcReady:
			while (0 == (ADC1->ISR & ADC_ISR_ADRDY)) {asm volatile("nop");}
			adc_deinit();
			m_state = wb_states::NoInit;
			break;
		case wb_states::WaitSensWarmup:
		case wb_states::MeasureChanStart:
			adc_deinit();
			m_state = wb_states::NoInit;
			break;
		case wb_states::MeasureChanWait:
			while (false == adc_done()) {asm volatile("nop");}
			adc_deinit();
			m_state = wb_states::NoInit;
			break;
		case wb_states::Calculate:
			m_state = wb_states::NoInit;
			break;
	}
	
	m_state = wb_states::NoInit;
}

barrel_err water_barrel::read_device(void)
{
	if (wb_states::NoInit == m_state)
	{
		return barrel_err::ERR_NOINIT;
	}
	
	if (wb_states::Idle == m_state)
	{
		m_state = wb_states::InitAdc;
	}
	
	return barrel_err::SUCCESS;
}

barrel_err water_barrel::read_progress(bool* doing)
{
	if (NULL == doing)
	{
		return barrel_err::ERR_ARGS;
	}
	
	if (wb_states::NoInit == m_state)
	{
		*doing = false;
		return barrel_err::ERR_NOINIT;
	}
	
	// if were here and its in idle then its doing nothing
	*doing = true;
	if (wb_states::Idle == m_state)
	{
		*doing = false;
	}
	
	return barrel_err::SUCCESS;
}

barrel_err water_barrel::read_level(uint32_t* level)
{
	if (NULL == level)
	{
		return barrel_err::ERR_ARGS;
	}
	
	if (wb_states::NoInit == m_state)
	{
		*level = 0;
		return barrel_err::ERR_NOINIT;
	}
	
	*level = m_water_level_mm;
	return barrel_err::SUCCESS;
}
	
barrel_err water_barrel::read_volume(uint32_t* volume)
{
	if (NULL == volume)
	{
		return barrel_err::ERR_ARGS;
	}
	
	if (wb_states::NoInit == m_state)
	{
		*volume = 0;
		return barrel_err::ERR_NOINIT;
	}
	
	*volume = m_water_volume_ml;
	return barrel_err::SUCCESS;
}
	
barrel_err water_barrel::read_volts(wb_readout what, uint16_t* volt)
{
	if (NULL == volt)
	{
		return barrel_err::ERR_ARGS;
	}
	
	if (wb_states::NoInit == m_state)
	{
		*volt = 0;
		return barrel_err::ERR_NOINIT;
	}
	
	switch (what)
	{
		case wb_readout::Sensor:
			*volt = m_values_mv[ARR_IDX_SENS];
			break;
		case wb_readout::VCC:
			*volt = m_values_mv[ARR_IDX_VCC];
			break;
	}
	
	return barrel_err::SUCCESS;
}
	
barrel_err water_barrel::read_adc(wb_readout what, uint16_t* bits)
{
	if (NULL == bits)
	{
		return barrel_err::ERR_ARGS;
	}
	
	if (wb_states::NoInit == m_state)
	{
		*bits = 0;
		return barrel_err::ERR_NOINIT;
	}
	
	switch (what)
	{
		case wb_readout::Sensor:
			*bits = m_values_adc[ARR_IDX_SENS];
			break;
		case wb_readout::VCC:
			*bits = m_values_adc[ARR_IDX_VCC];
			break;
	}
	
	return barrel_err::SUCCESS;
}

barrel_err water_barrel::process(void)
{
	bool expired;

	switch (m_state)
	{
		case wb_states::NoInit:
			break;
		
		case wb_states::Idle:
			break;
		
		case wb_states::InitAdc:
			m_pin_power_en->enable();
			m_timer_inst->start(&m_timer_idx, SENS_WARMUP_MS);
			adc_init();
			m_state = wb_states::WaitAdcKalib;
			break;
		
		case wb_states::WaitAdcKalib:
			if (0 == (ADC1->CR & ADC_CR_ADCAL))
			{
				ADC1->CR |= ADC_CR_ADEN;
				m_state = wb_states::WaitAdcReady;
			}
			break;
		
		case wb_states::WaitAdcReady:
			if (ADC1->ISR & ADC_ISR_ADRDY)
			{
				m_state = wb_states::WaitSensWarmup;
			}
			break;
		
		case wb_states::WaitSensWarmup:
			m_timer_inst->check_expiry(&m_timer_idx, &expired);
			if (expired)
			{
				m_readIdx = 0;
				m_state   = wb_states::MeasureChanStart;
			}
			break;
		
		case wb_states::MeasureChanStart:
			adc_start(adcChan[m_readIdx]);
			m_state = wb_states::MeasureChanWait;
			break;
		
		case wb_states::MeasureChanWait:
			if (adc_done())
			{
				adc_calcread(&m_values_mv[m_readIdx], &m_values_adc[m_readIdx], m_readIdx);
				m_readIdx++;
				if (m_readIdx <= ARR_IDX_MAX)
				{
					m_state = wb_states::MeasureChanStart;
				}
				else
				{
					m_pin_power_en->disable();
					adc_deinit();
					m_state = wb_states::Calculate;
				}
			}
			break;
		
		case wb_states::Calculate:
			calculate();
			m_state = wb_states::Idle;
			break;
	}
	
	return barrel_err::SUCCESS;
}

void water_barrel::can_sleep(bool *can_sleep)
{
	if (NULL == can_sleep)
	{
		return;
	}
	
	(*can_sleep) &= wb_states::Idle == m_state;
}

void water_barrel::adc_init(void)
{
	// TODO: clockon
	
	// Clear the ADRDY bit in ADC_ISR
	ADC1->ISR |= ADC_ISR_ADRDY;
	
	// disable all for reinit but not:
	// enable ADC vreg and wait until stable (TADCVREG_STUP L4 -> 20us)
	ADC1->CR = ADC_CR_ADVREGEN;
	//16MHz * 20us -> 320 nops  TODO: global wait us instead of this here
	uint32_t i;
	for (i = 0; i < 320; i++) {
		asm volatile("nop");
	}
	
	// single conversions, right alignment, 12bit
	ADC1->CFGR1 = 0;
	
	// no oversampling, clksrc = pclk/1
	// Over 32X, 5bit shift
	// Over 64X, 6bit shift
	#define PCLK1  0b11
	#define OVER32 0b100
	ADC1->CFGR2 = PCLK1 << ADC_CFGR2_CKMODE_Pos | 5 << ADC_CFGR2_OVSS_Pos | OVER32 << ADC_CFGR2_OVSR_Pos;
	
	// sampletime 160,5CLKs (16MHz and low current sensor output and voltage divider)
	#define SMP_160CLK 0b111
	ADC1->SMPR = SMP_160CLK << ADC_SMPR_SMP_Pos;
	
	// no vref, no presc
	ADC1_COMMON->CCR = 0;
	
	// calibrate in single ended mode (ADEN must be 0)
	ADC1->CR |= ADC_CR_ADCAL;
}

void water_barrel::adc_deinit()
{
	ADC1->CR |= ADC_CR_ADDIS;
	while(ADC1->CR & ADC_CR_ADDIS){asm volatile("nop");}
	ADC1->CR &= ~ADC_CR_ADVREGEN;
	// TODO: clockoff
}

void water_barrel::adc_start(uint8_t channel) 
{
	// set back end flag
	ADC1->ISR |= ADC_ISR_EOS;
	
	// set channel
	ADC1->CHSELR = 1 << channel;
	
	// enable conversion
	ADC1->CR |= ADC_CR_ADSTART;
}

bool water_barrel::adc_done(void)
{
	return 0 != (ADC1->ISR & ADC_ISR_EOS);
}

void water_barrel::adc_calcread(uint16_t *mv, uint16_t *adc, uint8_t chan_int) 
{
	
	uint16_t val = ADC1->DR;
	
	if (val <= m_config->adc_val_zero[chan_int]) {
		*mv  = 0;
		*adc = val;
		return;
	}
	
	val -= m_config->adc_val_zero[chan_int];
	uint32_t adc_5v = m_config->adc_val_full[chan_int] - m_config->adc_val_zero[chan_int];
	
	*mv  = ((uint32_t)val * (uint32_t)VOLT_FULLRANGE)/adc_5v;
	*adc = val;
}

void water_barrel::calculate(void)
{
	// Datasheet MPX5010: Vout = VCC * (0,09 * P + 0,04) [P in kPa = mm H2O]
	// P = (4/9) * (((25 * Vout)/VCC) - 1)
	if (0 != m_values_mv[ARR_IDX_VCC])
	{
		float level;
		level  = 25.0f * m_values_mv[ARR_IDX_SENS];
		level /= m_values_mv[ARR_IDX_VCC];
		level -= 1.0f;
		level *= 4.0f;
		level /= 9.0f;
		if (level < 0.0f)
		{
			m_water_level_mm = 0;
		}
		else
		{
			m_water_level_mm = level - m_config->pegel_offset;
		}
	}
	else
	{
		m_water_level_mm = 0;
	}
	
	// https://rechneronline.de/liter/eimer.php
	float f_water_level_cm = m_water_level_mm/10.0f;
	float f_barrel_dia_top_cm = m_config->barrel_dia_top/10.0f;
	float f_barrel_dia_bottom_cm = m_config->barrel_dia_bottom/10.0f;
	float f_barrel_height_cm = m_config->barrel_height/10.0f;
	
	float fillwidth;
	fillwidth = (f_barrel_dia_top_cm - f_barrel_dia_bottom_cm) * (f_water_level_cm/f_barrel_height_cm);
	fillwidth = f_barrel_dia_bottom_cm + fillwidth;
	
	float volume_ml;
	volume_ml = fillwidth*fillwidth + fillwidth*f_barrel_dia_bottom_cm + f_barrel_dia_bottom_cm*f_barrel_dia_bottom_cm;
	volume_ml = f_water_level_cm * M_PI*volume_ml/12.0f;
	m_water_volume_ml = volume_ml;
}