#ifndef WATER_BARREL_HPP
#define WATER_BARREL_HPP

#include "periph/i_pio.hpp"
#include "periph/i_timer_poll.hpp"

#include "i_sleepable.hpp"

#include <cstdint>
#include <cstddef>

enum class barrel_err {
	SUCCESS,
	ERR_NOINIT,
	ERR_ARGS,
};

enum class wb_readout {
	Sensor,
	VCC,
};

#define ARR_IDX_SENS  0
#define ARR_IDX_VCC   1
#define ARR_IDX_MAX   1

struct barrel_calibdata {
	uint16_t adc_val_zero[2];   //!< adc value for 0V
	uint16_t adc_val_full[2];   //!< adc value for 5V
	uint32_t pegel_offset;      //!< if the sensor is under the barrel [mm]
	uint32_t barrel_dia_bottom; //!< diameter at bottom of barrel [mm]
	uint32_t barrel_dia_top;    //!< diameter at top of barrel [mm]
	uint32_t barrel_height;     //!< height of barrel [mm]
	float    poly[3];           //!< polynom from mV to 0,1mm
};

class water_barrel : public i_sleepable {
public:

	water_barrel();
	
	//! \brief init of this instance
	//! \param[in] pin_power_en pin for power enable
	//! \param[in] config pointer to stored config data
	//! \return success or @ref barrel_err
	barrel_err init(pio_set* pin_power_en, const barrel_calibdata* config, i_timer_poll* timer_inst);
	
	void deinit(void);
	
	//! \brief readout the device, will work async, no temperture here
	//! \return success or @ref barrel_err
	barrel_err read_device(void);
	
	//! \brief checking that readout is in progress
	//! \param[out] doing true if measure is in progress
	//! \return success or @ref barrel_err
	barrel_err read_progress(bool* doing);
	
	//! \brief reads stored water level
	//! \param[out] level water level with LSB = 1mm
	//! \return success or @ref barrel_err
	barrel_err read_level(uint32_t* level);
	
	//! \brief reads stored water volume
	//! \param[out] volume water volume with LSB = 1ml
	//! \return success or @ref barrel_err
	barrel_err read_volume(uint32_t* volume);
	
	//! \brief reads voltage of sensor
	//! \param[in] what what to read
	//! \param[out] volt voltage with LSB = 1mV
	//! \return success or @ref barrel_err
	barrel_err read_volts(wb_readout what, uint16_t* volt);
	
	//! \brief reads adc bits of sensor
	//! \param[in] what what to read
	//! \param[out] bits bits of ADC
	//! \return success or @ref barrel_err
	barrel_err read_adc(wb_readout what, uint16_t* bits);
	
	//! \brief processing internals, call in main loop
	//! \return success or @ref barrel_err
	barrel_err process(void);
	
	virtual void can_sleep(bool *can_sleep);
	
private:

	enum class wb_states {
		NoInit,
		Idle,
		InitAdc,
		WaitAdcKalib,
		WaitAdcReady,
		WaitSensWarmup,
		MeasureChanStart,
		MeasureChanWait,
		Calculate,
	};

	void adc_init(void);
	void adc_deinit(void);
	
	void adc_start(uint8_t channel);
	bool adc_done(void);
	void adc_calcread(uint16_t *mv, uint16_t *adc, uint8_t chan_int);
	
	void calculate(void);
	
	pio_set* m_pin_power_en;
	i_timer_poll* m_timer_inst;
	
	const barrel_calibdata* m_config;
	
	tim_poll_obj m_timer_idx;
	
	wb_states m_state;
	uint8_t m_readIdx;
	
	uint16_t m_values_mv[2];
	uint16_t m_values_adc[2];
	
	uint32_t m_water_level_mm;
	uint32_t m_water_volume_ml;
};

#endif//WATER_BARREL_HPP