#include "config_data.hpp"

#include <cstddef>
#include <cstring>

#include "config_data_addr.hpp"

config_data::config_data() :
	m_eeprom_inst(NULL)
{
}
	
confdat_err config_data::init(eeprom* eeprom_inst, barrel_calibdata* barrel_calib, mb_setting* mb_settings)
{
	m_eeprom_inst = NULL;
	if (NULL == eeprom_inst)
	{
		return confdat_err::ERR_ARGS;
	}
	m_eeprom_inst = eeprom_inst;
	
	ee_err err;
	uint32_t readout;
	err = m_eeprom_inst->read(ADDR_FIRSTBLOOD, 4, &readout);
	if (ee_err::SUCCESS != err)
	{	
		return confdat_err::ERR_EXEC;
	}
	
	// its not a firstblood
	if (0 != readout)
	{
		return confdat_err::SUCCESS;
	}
	
	// its a firstboot
	confdat_err cerr;
	cerr = write_barrel_sizes(barrel_calib);
	if (confdat_err::SUCCESS != cerr)
	{
		return cerr;
	}
	cerr = write_barrel_calib(barrel_calib);
	if (confdat_err::SUCCESS != cerr)
	{
		return cerr;
	}
	cerr = write_mb_settings(mb_settings);
	if (confdat_err::SUCCESS != cerr)
	{
		return cerr;
	}
	
	// writeback
	readout = 0x1337;

	err = m_eeprom_inst->write(ADDR_FIRSTBLOOD, 4, &readout);
	if (ee_err::SUCCESS != err)
	{
		return confdat_err::ERR_EXEC;
	}

	return confdat_err::SUCCESS;
}
	
confdat_err config_data::write_barrel_sizes(barrel_calibdata* barrel_calib)
{
	if (NULL == m_eeprom_inst)
	{
		return confdat_err::ERR_INIT;
	}
	if (NULL == barrel_calib)
	{
		return confdat_err::ERR_ARGS;
	}
	
	ee_err err;
	
	err = m_eeprom_inst->write(ADDR_BARREL_DIA_BOT, 4, &barrel_calib->barrel_dia_bottom);
	if (ee_err::SUCCESS != err)
	{
		return confdat_err::ERR_EXEC;
	}
	err = m_eeprom_inst->write(ADDR_BARREL_DIA_TOP, 4, &barrel_calib->barrel_dia_top);
	if (ee_err::SUCCESS != err)
	{
		return confdat_err::ERR_EXEC;
	}
	err = m_eeprom_inst->write(ADDR_BARREL_HEIGHT,  4, &barrel_calib->barrel_height);
	if (ee_err::SUCCESS != err)
	{
		return confdat_err::ERR_EXEC;
	}
	
	return confdat_err::SUCCESS;
}
	
confdat_err config_data::write_barrel_calib(barrel_calibdata* barrel_calib)
{
	if (NULL == m_eeprom_inst)
	{
		return confdat_err::ERR_INIT;
	}
	if (NULL == barrel_calib)
	{
		return confdat_err::ERR_ARGS;
	}
	
	ee_err err;
		
	err = m_eeprom_inst->write(ADDR_BARREL_ADC_ZER, 4, barrel_calib->adc_val_zero);
	if (ee_err::SUCCESS != err)
	{
		return confdat_err::ERR_EXEC;
	}
	err = m_eeprom_inst->write(ADDR_BARREL_ADC_FUL, 4, barrel_calib->adc_val_full);
	if (ee_err::SUCCESS != err)
	{
		return confdat_err::ERR_EXEC;
	}
	err = m_eeprom_inst->write(ADDR_BARREL_OFFSET, 4, &barrel_calib->pegel_offset);
	if (ee_err::SUCCESS != err)
	{
		return confdat_err::ERR_EXEC;
	}
	err = m_eeprom_inst->write(ADDR_BARREL_POLY0, 4, &barrel_calib->poly[0]);
	if (ee_err::SUCCESS != err)
	{
		return confdat_err::ERR_EXEC;
	}
	err = m_eeprom_inst->write(ADDR_BARREL_POLY1, 4, &barrel_calib->poly[1]);
	if (ee_err::SUCCESS != err)
	{
		return confdat_err::ERR_EXEC;
	}
	err = m_eeprom_inst->write(ADDR_BARREL_POLY2, 4, &barrel_calib->poly[2]);
	if (ee_err::SUCCESS != err)
	{
		return confdat_err::ERR_EXEC;
	}
	
	return confdat_err::SUCCESS;
}
	
confdat_err config_data::write_mb_settings(mb_setting* mb_settings)
{
	if (NULL == m_eeprom_inst)
	{
		return confdat_err::ERR_INIT;
	}
	if (NULL == mb_settings)
	{
		return confdat_err::ERR_ARGS;
	}
	
	ee_err err;
	
	
	err = m_eeprom_inst->write(ADDR_MODBUS, 4, mb_settings);
	if (ee_err::SUCCESS != err)
	{
		return confdat_err::ERR_EXEC;
	}
	
	return confdat_err::SUCCESS;
}
	
confdat_err config_data::read(barrel_calibdata* barrel_calib)
{
	if (NULL == m_eeprom_inst)
	{
		return confdat_err::ERR_INIT;
	}
	if (NULL == barrel_calib)
	{
		return confdat_err::ERR_ARGS;
	}
	
	bool errfree = true;
	readout(ADDR_BARREL_DIA_BOT, &barrel_calib->barrel_dia_bottom, 4, 4, &errfree);
	readout(ADDR_BARREL_DIA_TOP, &barrel_calib->barrel_dia_top,    4, 4, &errfree);
	readout(ADDR_BARREL_HEIGHT , &barrel_calib->barrel_height,     4, 4, &errfree);
	readout(ADDR_BARREL_ADC_ZER, &barrel_calib->adc_val_zero,      4, 4, &errfree);
	readout(ADDR_BARREL_ADC_FUL, &barrel_calib->adc_val_full,      4, 4, &errfree);
	readout(ADDR_BARREL_OFFSET,  &barrel_calib->pegel_offset,      4, 4, &errfree);
	readout(ADDR_BARREL_POLY0,   &barrel_calib->poly[0],           4, 4, &errfree);
	readout(ADDR_BARREL_POLY1,   &barrel_calib->poly[1],           4, 4, &errfree);
	readout(ADDR_BARREL_POLY2,   &barrel_calib->poly[2],           4, 4, &errfree);
	if (false == errfree)
	{
		return confdat_err::ERR_EXEC;
	}
	
	return confdat_err::SUCCESS;
}
	
confdat_err config_data::read(mb_setting* mb_settings)
{
	if (NULL == m_eeprom_inst)
	{
		return confdat_err::ERR_INIT;
	}
	if (NULL == mb_settings)
	{
		return confdat_err::ERR_ARGS;
	}
	
	bool errfree = true;
	readout(ADDR_MODBUS, mb_settings, 4, sizeof(*mb_settings), &errfree);
	if (false == errfree)
	{
		return confdat_err::ERR_EXEC;
	}
	return confdat_err::SUCCESS;
}

void config_data::readout(uint32_t addr, void* data, uint8_t sizeee, uint8_t sizecopy, bool* errfree)
{
	ee_err err;
	uint32_t readout;
	err = m_eeprom_inst->read(addr, sizeee, (void*)&readout);
	if (ee_err::SUCCESS != err)
	{	
		*errfree = false;
		return;
	}
	memcpy(data, &readout, sizecopy);
}