#include <stdint.h>
#include <cstring>

#include "cmsis_includer.h"

#include "periph/i_pio.hpp"
#include "periph/pio.hpp"
#include "periph/i_timer_irq.hpp"
#include "periph/timer_irq_base.hpp"
#include "periph/i_timer_poll.hpp"
#include "periph/timer_poll_base.hpp"
#include "periph/mb_uart32_de_irq.hpp"
#include "periph/i_i2c.hpp"
#include "periph/i2c_l4.hpp"
#include "periph/eeprom.hpp"

#include "device/lm75_sleep.hpp"
#include "device/water_barrel.hpp"

#include "hal/config_data.hpp"

#include "modbus/modbus.hpp"

#include "mb_callb_coil.hpp"
#include "mb_callb_hold.hpp"
#include "mb_callb_inreg.hpp"

#include "cpu_types.h"

// get constants from linkerscript
extern void (*_flash_start []) (void) __attribute__((weak));

// Pins
const pio_pin pin_i2c1_sda = {.port = GPIOB, .nbr =  7, .mode = AF_OD,     .speed = GPIO_HS, .alt_func = 1};
const pio_pin pin_i2c1_scl = {.port = GPIOB, .nbr =  6, .mode = AF_OD,     .speed = GPIO_HS, .alt_func = 1};

const pio_pin pin_uart2_tx = {.port = GPIOA, .nbr =  2, .mode = AF_PP,     .speed = GPIO_HS, .alt_func = 4};
const pio_pin pin_uart2_rx = {.port = GPIOA, .nbr = 15, .mode = AF_PP,     .speed = GPIO_HS, .alt_func = 4};
const pio_pin pin_uart2_de = {.port = GPIOA, .nbr = 12, .mode = GPO_PP,    .speed = GPIO_HS, .alt_func = 0};

const pio_pin pin_sens_en  = {.port = GPIOA, .nbr =  3, .mode = GPO_PP,    .speed = GPIO_MS, .alt_func = 0};
const pio_pin pin_lm75_en  = {.port = GPIOB, .nbr =  5, .mode = GPO_PP,    .speed = GPIO_MS, .alt_func = 0};

const pio_pin pin_adc_in0  = {.port = GPIOA, .nbr =  0, .mode = GP_ANALOG, .speed = GPIO_HS, .alt_func = 0};
const pio_pin pin_adc_in1  = {.port = GPIOA, .nbr =  1, .mode = GP_ANALOG, .speed = GPIO_HS, .alt_func = 0};

const pio_pin pin_debug    = {.port = GPIOA, .nbr =  4, .mode = GPO_PP,    .speed = GPIO_MS, .alt_func = 0};

static pio pio_inst;
static pio_set pin_power_lm75;
static pio_set pin_power_sens;
static pio_set pin_send;
static pio_set pin_debug_inst;

static timer_irq_base   timer_modbus_inst;
static timer_poll_base  timer_poll_inst;
static mb_uart32_de_irq mb_uart_inst;
static i2c_l4 i2c1_inst;
static eeprom eepom_inst;

static config_data confdat_inst;

static lm75_sleep lm75;
static water_barrel water;

static modbus modbusInst;
static mb_callback_coils mb_coils;
static mb_callback_hold  mb_hold;
static mb_callback_inreg mb_inreg;

extern "C" {
	extern void asm_jumper(uint32_t pc, uint32_t sp);
}

static const pio_pin_set conf_power_pin_lm75 = {
	.pin = &pin_lm75_en, 
	.enableIsHigh = true, // Pin directly to LM75 VCC
};

static const pio_pin_set conf_power_pin_sens = {
	.pin = &pin_sens_en, 
	.enableIsHigh = false, // PFET for 5V
};

static const pio_pin_set conf_pin_send = {
	.pin = &pin_uart2_de, 
	.enableIsHigh = true,
};

static const pio_pin_set conf_pin_debug_inst = {
	.pin = &pin_debug, 
	.enableIsHigh = true,
};

static const timer_irq_base_config timer_mb_conf = {
	.timer   = TIM2,
	.irq_nbr = TIM2_IRQn, // 15
};

static timer_irq_base_sleep timer_mb_sleep = {
	.tim_sl_reg = &RCC->APB1SMENR,
	.tim_sl_bit = RCC_APB1SMENR_TIM2SMEN,
};

static const timer_poll_base_config timer_poll_conf = {
	.timer   = TIM21,
};

static const mb_uart_config mb_uart_config = {
	.usart      = USART2,
	.irq_nbr    = USART2_IRQn, // 28
	.timer_inst = &timer_modbus_inst,
	.send_en    = &pin_send,
};

// filled with default settings
static barrel_calibdata conf_water = {
	.adc_val_zero = {0, 5},
	.adc_val_full = {3845, 3833},
	.pegel_offset = 0,
	.barrel_dia_bottom = 1000,
	.barrel_dia_top = 1000,
	.barrel_height = 1000,
	.poly = {0.0, 1.0, 0.0},  
};

// filled with default settings
static mb_setting mb_settings = {
	.address  = MB_ADDRESS_DEFAULT,
	.baudrate = mb_baudrate::b115200,
	.parity   = mb_parmode::NOP_2S,
};

__attribute__((section(".cputypes")))
__attribute__((used))
static const cpu_types_supported cpu = {
	.fw_vers  = 0 << 24 | 2 << 16 | 0 << 8 | 7,
	.num      = 1,
	.typesize = sizeof(cpu_types),
	.types    = {cpu_type_STM32L010K8},
};

extern "C" int main (void) {

	// switch to HSI16, its precise enough for uart
	RCC->CR |= RCC_CR_HSION;
	while (0 == (RCC->CR & RCC_CR_HSIRDY)) {asm volatile("nop");}
	RCC->CFGR |= 1 << RCC_CFGR_SW_Pos;
	while (0 == (RCC->CR & RCC_CFGR_SW)) {asm volatile("nop");}
	
	// set vector table direct to _OUR_ flash start instead of remapped 0x0 area that shows to the BL
	SCB->VTOR = (uint32_t)_flash_start;
	__enable_irq();
	
	// enable all needed periph clocks
	RCC->IOPENR  |= RCC_IOPENR_IOPAEN | RCC_IOPENR_IOPBEN | RCC_IOPENR_IOPCEN;
	RCC->APB1ENR |= RCC_APB1ENR_I2C1EN | RCC_APB1ENR_USART2EN | RCC_APB1ENR_TIM2EN;
	RCC->APB2ENR |= RCC_APB2ENR_DBGEN | RCC_APB2ENR_ADCEN | RCC_APB2ENR_TIM21EN | RCC_APB2ENR_SYSCFGEN; // SYSCFG -> EXTI
	
	// not needed periph in sleep mode (only disable TIM2 if no RXing is ongoing! -> handled in timer driver)
	RCC->APB1SMENR &= ~(RCC_APB1SMENR_I2C1SMEN /*| RCC_APB1SMENR_TIM2SMEN*/);
	RCC->APB2SMENR &= ~(RCC_APB2SMENR_ADCSMEN | RCC_APB2SMENR_TIM21SMEN);
	
	// init GPIOs
	pio_inst.init(&pin_uart2_tx);
	pio_inst.init(&pin_uart2_rx);
	pio_inst.init(&pin_uart2_de);
	
	pio_inst.init(&pin_adc_in0);
	pio_inst.init(&pin_adc_in1);
	
	pio_inst.init(&pin_debug);
	
	// first enable I2C periph and after that the pins
	// or else there are some glitches in the line that confuses the slaves
	// source: internet
	i2c1_inst.init(I2C1, 16000000, 100000);
	pio_inst.init(&pin_i2c1_sda);
	pio_inst.init(&pin_i2c1_scl);
	
	// init wrapped GPIOs
	pin_power_lm75.init(&pio_inst, &conf_power_pin_lm75);
	pin_power_sens.init(&pio_inst, &conf_power_pin_sens);
	pin_send.init(&pio_inst, &conf_pin_send);
	pin_debug_inst.init(&pio_inst, &conf_pin_debug_inst);
	
	// init timers
	timer_modbus_inst.init(&timer_mb_conf, &timer_mb_sleep, NULL);
	timer_poll_inst.init(&timer_poll_conf, &pin_debug_inst);
	
	// config data from eeprom
	confdat_inst.init(&eepom_inst, &conf_water, &mb_settings);
	/*confdat_inst.write_barrel_sizes(&conf_water);
	confdat_inst.write_barrel_calib(&conf_water);
	confdat_inst.write_mb_settings(&mb_settings);*/
	confdat_inst.read(&conf_water);
	confdat_inst.read(&mb_settings);
	
	// init sensors
	lm75.init(&i2c1_inst, &pin_power_lm75, &timer_poll_inst);
	water.init(&pin_power_sens, &conf_water, &timer_poll_inst);
	
	// init modbus and its callbacks
	mb_uart_inst.init(&mb_uart_config);
	modbusInst.init(&mb_uart_inst, mb_settings.baudrate, mb_settings.parity);
	modbusInst.set_address(mb_settings.address);
	
	mb_coils.init(&modbusInst, &mb_settings, &lm75, &water, &confdat_inst, &conf_water);
	modbusInst.register_callback(mb_callback_type::coil, &mb_coils);
	
	mb_hold.init(&conf_water, &modbusInst, &mb_settings);
	modbusInst.register_callback(mb_callback_type::register_hold, &mb_hold);
	
	mb_inreg.init(&lm75, &water);
	modbusInst.register_callback(mb_callback_type::register_input, &mb_inreg);
 
	while(1) 
	{
		lm75.process();
		water.process();
		modbusInst.process();
		
		// restarting to BL if modbus has send answer and its flagged
		bool load_bl = true;
		load_bl &= mb_coils.bl_is_flagged();
		modbusInst.can_sleep(&load_bl); // only fires if answer was send
		if (load_bl)
		{
			// deinit everything that can throw IRQs and/or is not stateless
			water.deinit();
			lm75.deinit();
			i2c1_inst.deinit();
			mb_uart_inst.deinit();
			timer_modbus_inst.deinit();
			pio_inst.deinit(&pin_adc_in0);
			pio_inst.deinit(&pin_adc_in1);
			pio_inst.deinit(&pin_debug);
			pio_inst.deinit(&pin_i2c1_sda);
			pio_inst.deinit(&pin_i2c1_scl);
			__disable_irq();
			GPIOB->PUPDR = 0x13370042; // Bootloader request
			volatile uint32_t* vectors = (volatile uint32_t*)0x08000000;
			uint32_t pc = vectors[1];
			uint32_t sp = vectors[0];
			asm_jumper(pc, sp);
		}
		
		bool can_sleep = true;
		water.can_sleep(&can_sleep);
		lm75.can_sleep(&can_sleep);
		modbusInst.can_sleep(&can_sleep);
		
		/*
		Entering Low-power sleep mode
		To enter Low-power sleep mode, proceed as follows:
		1. The Flash memory can be switched off by using the control bits (SLEEP_PD in the
		FLASH_ACR register. This reduces power consumption but increases the wake-up
		time.
		2. Each digital IP clock must be enabled or disabled by using the RCC_APBxENR and
		RCC_AHBENR registers.
		3. The frequency of the system clock must be decreased.
		4. The regulator is forced in low-power mode by software (LPSDSR bits set).
		5. Follow the steps described in Section 6.3.5: Entering low-power mode.
		*/
		if (can_sleep)
		{
			FLASH->ACR |= FLASH_ACR_SLEEP_PD;
			PWR->CR |= PWR_CR_LPSDSR;// low power
			asm("wfi"); // sleep mode
		}
	}
	
	return 0;
}

extern "C" void TIM2_IRQHandler(void)
{
	timer_modbus_inst.irq();
}

extern "C" void USART2_IRQHandler(void)
{
	mb_uart_inst.irq();
}