#include "eeprom.hpp"

#include <cstddef>

#include "cmsis_includer.h"

#define EEPROM_SIZE       256
#define EEPROM_END_ADDR   (EEPROM_START_ADDR + EEPROM_SIZE - 1)
#define EEPROM_PAGE_SIZE  4

eeprom::eeprom()
{
}

ee_err eeprom::erase(uint32_t addr_start, uint32_t size)
{
	ee_err err = check_input(addr_start, size);
	if (ee_err::SUCCESS != err)
	{
		return err;
	}
	
	if (is_busy())
	{
		return ee_err::BUSY;
	}
	
	unlock();
	erase_flags();
	
	volatile uint32_t* dst = (volatile uint32_t*)addr_start;
	
	for (; size > 0; size -= 4)
	{
		if (*dst != 0x00000000)
		{
			*dst  = 0x00000000;
			while(is_busy())
			{
				asm volatile("nop");
			}
		}
		dst++;
	}
	
	lock();
	
	return err;
}

ee_err eeprom::write(uint32_t addr_start, uint32_t size, void* data)
{
	ee_err err = check_input(addr_start, size);
	if (ee_err::SUCCESS != err)
	{
		return err;
	}
	
	volatile uint32_t* dst = (volatile uint32_t*)addr_start;
	volatile uint32_t* src = (volatile uint32_t*)data;
	
	if (*dst == *src)
	{
		// dont write the same thing twice
		return ee_err::SUCCESS;
	}
	
	if (is_busy())
	{
		return ee_err::BUSY;
	}
	
	unlock();
	erase_flags();
	
	for (; size > 0; size -= 4)
	{
		if (*dst != *src)
		{
			*dst  = *src;
			while(is_busy())
			{
				asm volatile("nop");
			}
		}
		src++;
		dst++;
	}
	
	lock();
	
	return err;
}
	
ee_err eeprom::read(uint32_t addr_start, uint32_t size, void* data)
{
	ee_err err = check_input(addr_start, size);
	if (ee_err::SUCCESS != err)
	{
		return err;
	}
	
	volatile uint32_t* src = (volatile uint32_t*)addr_start;
	volatile uint32_t* dst = (volatile uint32_t*)data;
	
	for (; size > 0; size -= 4)
	{
		*dst = *src;
		src++;
		dst++;
	}
	
	return err;
}

bool eeprom::is_busy(void)
{
	return 0 != (FLASH->SR & FLASH_SR_BSY);
}

bool eeprom::is_locked()
{
	return 0 != (FLASH->PECR & FLASH_PECR_PELOCK);
}
	
void eeprom::unlock()
{
	if (is_locked())
	{
		__disable_irq();
		FLASH->PEKEYR = 0x89ABCDEF;
		FLASH->PEKEYR = 0x02030405;
		__enable_irq();	
		while(is_locked())
		{
			asm volatile("nop");
		}
	}
}
	
void eeprom::lock()
{
	FLASH->PECR |= FLASH_PECR_PELOCK;
}

bool eeprom::is_in_space(uint32_t addr_start, uint32_t size)
{
	return ((addr_start >= EEPROM_START_ADDR) && ((addr_start + size - 1) <= EEPROM_END_ADDR));
}
	
bool eeprom::is_aligned(uint32_t addr_start, uint32_t size)
{
	return ((0 == addr_start%4) && (0 == size%4));
}

ee_err eeprom::check_input(uint32_t addr_start, uint32_t size)
{
	if (false == is_in_space(addr_start, size))
	{
		return ee_err::ERR_SPACE;
	}
	if (false == is_aligned(addr_start, size))
	{
		return ee_err::ERR_ALIGN;
	}
	return ee_err::SUCCESS;
}

void eeprom::erase_flags(void)
{
	FLASH->SR |= FLASH_SR_FWWERR | FLASH_SR_NOTZEROERR | FLASH_SR_RDERR | FLASH_SR_OPTVERR | FLASH_SR_SIZERR | FLASH_SR_PGAERR | FLASH_SR_WRPERR;
}