#ifndef EEPROM_HPP
#define EEPROM_HPP

#include <cstdint>

#define EEPROM_START_ADDR 0x08080000

enum class ee_err {
	SUCCESS,
	BUSY,
	ERR_ARGS,
	ERR_INIT,
	ERR_ALIGN, //!< not aligned
	ERR_SPACE, //!< not in eeprom address space
};

class eeprom {
public:

	eeprom();
	
	//! \brief erases eeprom space
	//! \param[in] addr_start start address; has to be in eeprom space and page aligned
	//! \param[in] size size to erase; has to be page aligned
	//! \return eeprom error code @ref ee_err
	ee_err erase(uint32_t addr_start, uint32_t size);
	
	//! \brief writes to eeprom space
	// will not write of a word is the same
	//! \param[in] addr_start start address; has to be in eeprom space and page aligned
	//! \param[in] size size to write; has to be page aligned
	//! \param[in] data data to be written
	//! \return eeprom error code @ref ee_err
	ee_err write(uint32_t addr_start, uint32_t size, void* data);
	
	//! \brief reads from eeprom space
	//! \param[in] addr_start start address; has to be in eeprom space and page aligned
	//! \param[in] size size to read; has to be page aligned
	//! \param[out] data data to be read into this memory space
	//! \return eeprom error code @ref ee_err
	ee_err read(uint32_t addr_start, uint32_t size, void* data);

	//! \brief polls if eeprom is busy
	//! \return true if busy
	bool is_busy(void);

private:
	
	//! \brief checks that eeprom is locked
	//! \return true if locked, else false
	bool is_locked();
	
	//! \brief unlocks eeprom
	void unlock();
	
	//! \brief locks eeprom
	void lock();
	
	bool is_in_space(uint32_t addr_start, uint32_t size);
	
	bool is_aligned(uint32_t addr_start, uint32_t size);
	
	ee_err check_input(uint32_t addr_start, uint32_t size);
	
	void erase_flags(void);
};

#endif // EEPROM_HPP