
#include "i2c_l4.hpp"

#include <cstddef>

#define I2C_CHUNKSIZE 255

i2c_l4::i2c_l4() :
	i2c(NULL) {
}

i2c_err i2c_l4::init(I2C_TypeDef *i2c_in, uint32_t clock_bus, uint32_t clock_i2c) {

	(void)clock_bus;
	(void)clock_i2c;
	
	i2c = i2c_in;
	
	i2c->CR1 = 0; // all off and status reset
	i2c->CR2 = 0; // 7Bit address
	
	i2c->TIMINGR = 0x00303D5B; // TODO: not typing from mx
	
	// enable at last
	i2c->CR1 |= I2C_CR1_PE;
	
	return i2c_err::SUCCESS;
}

void i2c_l4::deinit(void)
{
	if (NULL == i2c)
	{
		return;
	}
	
	i2c->CR1 &= ~I2C_CR1_PE;
	i2c = NULL;
}

// address left justified (LSB is RW)
i2c_err i2c_l4::write(uint8_t address, uint8_t *data, uint32_t len, bool stopafter) {
	
	// Master communication initialization (address phase)
	setaddress(address, i2cdir::send, stopafter);
	
	i2c_err err = i2c_err::SUCCESS;
	bool start = true;
	uint32_t sent = 0;
	
	// i2c on L431 works with 255 byte chunks
	while (len && (err == i2c_err::SUCCESS)) {
		uint32_t chunksize   = len;
		bool stopafter_chunk = stopafter && (len < I2C_CHUNKSIZE);
		err   = write_chunk(data + sent, &chunksize, start, stopafter_chunk, false);
		start = false; // only start once
		len  -= chunksize;
		sent += chunksize;
	}
	
	return err;
}

i2c_err i2c_l4::write(uint8_t address, uint8_t *command, uint8_t clen, uint8_t *data, uint32_t len, bool stopafter) {
	
	// Master communication initialization (address phase)
	setaddress(address, i2cdir::send, false); // stopafter is evaluated later
	
	// send command first
	i2c_err err;
	uint32_t sent = 0;
	uint32_t commandlen = clen;
	err = write_chunk(command, &commandlen, true, false, true); // start, end, reload
	
	// i2c on L431 works with 255 byte chunks
	while (len && (err == i2c_err::SUCCESS)) {
		uint32_t chunksize   = len;
		bool stopafter_chunk = stopafter && (len < I2C_CHUNKSIZE);
		err   = write_chunk(data + sent, &chunksize, false, stopafter_chunk, false);
		len  -= chunksize;
		sent += chunksize;
	}
	
	return err;
}

i2c_err i2c_l4::read(uint8_t address, uint8_t *command, uint8_t clen, uint8_t *data, uint32_t len, bool stopafter) {
	
	// Master communication initialization (address phase)
	setaddress(address, i2cdir::send, false); // stopafter is evaluated later
	
	// send command first
	i2c_err err;
	uint32_t received = 0;
	uint32_t commandlen = clen;
	err = write_chunk(command, &commandlen, true, true, false); // start, end, reload
	
	// restart 
	setaddress(address, i2cdir::receive, false); // stopafter is evaluated later
	bool restart = true;
	
	// i2c on L431 works with 255 byte chunks
	while (len && (err == i2c_err::SUCCESS)) {
		uint32_t chunksize   = len;
		bool stopafter_chunk = stopafter && (len < I2C_CHUNKSIZE);
		err   = read_chunk(data + received, &chunksize, restart, stopafter_chunk, false);
		restart   = false;
		len      -= chunksize;
		received += chunksize;
	}
	
	return err;
}

i2c_err i2c_l4::write_chunk(uint8_t *data, uint32_t *chunksize, bool start, bool stopafter, bool reload) {
	
	i2c_err err = i2c_err::SUCCESS;
	
	uint32_t waitfor_bit = start_chunk(chunksize, reload);
	
	if (stopafter) {
		// autoend 0 -> for a restart; 1 -> stop
		// Note: This bit has no effect in slave mode or when the RELOAD bit is set.
		i2c->CR2 |= I2C_CR2_AUTOEND;
	}
	
	if (start) {
		// The user must then set the START bit in I2C_CR2 register. Changing all the above bits is
		// not allowed when START bit is set.
		i2c->CR2 |= I2C_CR2_START;
	}
	
	uint32_t i;
	for (i = 0; i < *chunksize; i++) {
		err = sentpoll();
		if (err != i2c_err::SUCCESS) {
			return err;
		}
		i2c->TXDR = data[i];
	}
	
	// override for stop flag if (automatic) stopped after this byte
	if (stopafter) {
		waitfor_bit = I2C_ISR_STOPF;
	}
	
	// wait for complete
	while(1) {
		if (i2c->ISR & waitfor_bit) {
			break;
		}
	}
	
	if (stopafter) {
		i2c->ICR = I2C_ICR_STOPCF;
	}
	
	return err;
}

i2c_err i2c_l4::read_chunk(uint8_t *data, uint32_t *chunksize, bool start, bool stopafter, bool reload) {
	
	i2c_err err = i2c_err::SUCCESS;
	
	uint32_t waitfor_bit = start_chunk(chunksize, reload);
	
	if (stopafter) {
		// autoend 0 -> for a restart; 1 -> stop
		// Note: This bit has no effect in slave mode or when the RELOAD bit is set.
		i2c->CR2 |= I2C_CR2_AUTOEND;
	}
	
	if (start) {
		// The user must then set the START bit in I2C_CR2 register. Changing all the above bits is
		// not allowed when START bit is set.
		i2c->CR2 |= I2C_CR2_START;
	}
	
	uint32_t i;
	for (i = 0; i < *chunksize; i++) {
		err = receivedpoll();
		if (err != i2c_err::SUCCESS) {
			return err;
		}
		data[i] = i2c->RXDR;
	}
	
	// override for stop flag if (automatic) stopped after this byte
	if (stopafter) {
		waitfor_bit = I2C_ISR_STOPF;
	}
	
	// wait for complete
	while(1) {
		if (i2c->ISR & waitfor_bit) {
			break;
		}
	}
	
	if (stopafter) {
		i2c->ICR = I2C_ICR_STOPCF;
	}
	
	return err;
}

uint32_t i2c_l4::start_chunk(uint32_t *chunksize, bool reload) {

	uint32_t reg = i2c->CR2;
	
	uint32_t waitfor_bit;
	if (*chunksize > I2C_CHUNKSIZE) {
		*chunksize = I2C_CHUNKSIZE;
		reg |= I2C_CR2_RELOAD; // do not stop after this 255 bytes
		waitfor_bit = I2C_ISR_TCR;
	} else if (reload) {
		reg |= I2C_CR2_RELOAD; // do not stop after this <255 bytes
		waitfor_bit = I2C_ISR_TCR;
	} else {
		reg &= ~I2C_CR2_RELOAD;
		waitfor_bit = I2C_ISR_TC;
	}
	
	// The number of bytes to be transferred: NBYTES[7:0]. If the number of bytes is equal to
	// or greater than 255 bytes, NBYTES[7:0] must initially be filled with 0xFF.
	reg &= ~I2C_CR2_NBYTES_Msk;
	reg |= (*chunksize) << I2C_CR2_NBYTES_Pos;
	
	i2c->CR2 = reg;
	
	return waitfor_bit;
}

i2c_err i2c_l4::setaddress(uint8_t address, i2cdir dir, bool stopafter) {
	
	// In order to initiate the communication, the user must program the following parameters for
	// the addressed slave in the I2C_CR2 register:
	uint32_t reg = i2c->CR2;
	
	// Addressing mode (7-bit or 10-bit): ADD10
	// Slave address to be sent: SADD[9:0]
	reg &= ~I2C_CR2_SADD_Msk;
	reg |= address;
	
	// Transfer direction: RD_WRN
	if (i2cdir::receive == dir) {
		reg |=  I2C_CR2_RD_WRN;
	} else {
		reg &= ~I2C_CR2_RD_WRN; // write
	}
	
	reg &= ~I2C_CR2_AUTOEND;
	if (stopafter) {
		// autoend 0 -> for a restart; 1 -> stop
		// Note: This bit has no effect in slave mode or when the RELOAD bit is set.
		reg |= I2C_CR2_AUTOEND;
	}
	i2c->CR2 = reg;
	
	return i2c_err::SUCCESS;
}

i2c_err i2c_l4::sentpoll() {
	
	while(1) {
		uint32_t sr = i2c->ISR;
		
		// ISR NACKF = 1 -> fail
		if (sr & I2C_ISR_NACKF) {
			i2c->ICR = I2C_ICR_NACKCF;
			return i2c_err::NACKED;
		}
		
		// ISR TXIS = 1 -> OK
		if (sr & I2C_ISR_TXIS) {
			return i2c_err::SUCCESS;
		}
	}
}

i2c_err i2c_l4::receivedpoll() {
	
	while(1) {
		uint32_t sr = i2c->ISR;
		
		// ISR NACKF = 1 -> fail
		if (sr & I2C_ISR_NACKF) {
			i2c->ICR = I2C_ICR_NACKCF;
			return i2c_err::NACKED;
		}
		
		// ISR RXNE = 1 -> OK
		if (sr & I2C_ISR_RXNE) {
			return i2c_err::SUCCESS;
		}
	}
}