#ifndef I_DMA_HPP
#define I_DMA_HPP

#include <cstdint>

//! \brief DMA errs
enum class dma_err {
	SUCCESS,		//!< no errors
	ERR_INV_ARGS,	//!< invalid arguments
	ERR_EXEC,		//!< error during execution
	ERR_USED,		//!< not available, already used
};

//! \brief dma errors
enum class dma_errflag : uint32_t {
	NONE   = 0,			//!< no error
	DONE   = 1 << 0,	//!< dma done
	COMMON = 1 << 1,	//!< min 1 error is set or the DMA only has one general error flag
	HALF   = 1 << 2,	//!< half transfer done
};

//! \brief interface of a DMA device
class i_dma_device {
public:
	constexpr i_dma_device(){}
};

//! \brief interface class of a DMA allocation
class i_dma_alloc {
public:
	constexpr i_dma_alloc(){}
};

enum class dma_prio {
	LOW   = 0,
	MED   = 1,
	HIGH  = 2,
	VHIGH = 3,
};

enum class dma_size {
	b8  = 0,
	b16 = 1,
	b32 = 2,
};

enum class dma_dir {
	PER2MEM = 0,
	MEM2PER = 1,
};

//! \brief void callback_func(uint32_t err_flags, void* userarg)
typedef void (*dma_callb)(uint32_t, void*);

struct dma_subconfig {
	dma_size  size; //!< element
	bool      incr; //!< increment address
	uint32_t *addr; //!< (start) address 
};

//! \brief Interface of a DMA Config
struct dma_config {
	bool          mem2mem;	//!< dont wait for handshakes
	dma_prio      prio;     //!< prio of this transfer
	dma_subconfig mem;		//!< config of memory part
	dma_subconfig per;		//!< config of periph part
	bool          circular; //!< circular mode if true
	dma_dir       trans_dir;//!< transfer direction
	bool          irq_half; //!< IRQ on half transfer
	bool          irq_err;  //!< IRQ on error
	uint16_t      num_trans;//!< number of transfers
	dma_callb     callb;    //!< done and err (irq) callback; NULL -> no IRQ
    void*         userarg;  //!< userarg for callback
};

class i_dma {
public:

	i_dma() {}
    
	//! \brief configures the given channel on this DMA for a future transfer
	//! \param[in] alloc pointer to alloc config
	//! \param[in] config pointer to config parameters structure of transfer
	//! \return SUCCESS or @ref dma_err
	virtual dma_err config(const i_dma_alloc *alloc, dma_config *config) = 0;
    
	//! \brief reconfigures the given channel on this DMA for a future transfer
	//! \details only sets values that are changed by hw during last transfer
	//! \param[in] alloc pointer to alloc config
	//! \param[in] config pointer to config parameters structure of transfer
	//! \return SUCCESS or @ref dma_err
	virtual dma_err reconfig(const i_dma_alloc *alloc, dma_config *config) = 0;
    
	//! \brief reconfigures the given channel on this DMA for a future transfer
	//! \details additional sets mem and size new
	//! \param[in] alloc pointer to alloc config
	//! \param[in] config pointer to config parameters structure of transfer
	//! \param[in] addrMem memory address do dma to/from
	//! \param[in] sizeTransfer size of the future transfer
	//! \return SUCCESS or @ref dma_err
	virtual dma_err reconfig(const i_dma_alloc *alloc, dma_config *config,
							 uint32_t addrMem, uint32_t sizeTransfer) = 0;
    
	//! \brief starts the given channel on this DMA
	//! \param[in] alloc pointer to alloc config
	//! \return SUCCESS or @ref dma_err
	virtual dma_err start(const i_dma_alloc *alloc) = 0;
    
	//! \brief aborts (stops transfer) for the given channel on this DMA
	//! \param[in] alloc pointer to alloc config
	//! \param[out] recBytes pointer to store transferred elements until abort
	//! \return SUCCESS or @ref dma_err
	virtual dma_err abort(const i_dma_alloc *alloc, uint32_t *recBytes) = 0;
    
	//! \brief flushes the fifo (do after abort) of the given channel on this DMA
	//! \param[in] alloc pointer to alloc config
	//! \return SUCCESS or @ref dma_err
	virtual dma_err flush(const i_dma_alloc *alloc) = 0;
    
	//! \brief poll DMA without IRQ until done on the given channel on this DMA
	//! \param[in] alloc pointer to alloc config
	//! \return SUCCESS or @ref dma_err
	virtual dma_err poll(const i_dma_alloc *alloc) = 0;
    
	//! \brief gets remaining bytes of this dma channel (can be running or after done)
	//! \param[in] alloc pointer to alloc config
	//! \param[out] remain pointer to store transferred elements until abort
	//! \return SUCCESS or @ref dma_err
	virtual dma_err get_remaining_elems(const i_dma_alloc *alloc, uint32_t *remain) = 0;
};

#endif//I_DMA_HPP