#ifndef I_PIO_HPP
#define I_PIO_HPP

#include <cstdint>

//! \brief structure for one PIO pin (forward declared)
struct pio_pin;

//! \brief structure for one PIO pin setter
struct pio_pin_set {
	const pio_pin* pin; 
	bool enableIsHigh;
};

enum class pioerr {
	SUCCESS,
	ERR_ARGS,
	ERR_INIT,
};

class i_pio {
public:
	
	//! \brief init a single pin
	//! \param pin pin to init
	//! \return @ref pioerr
	virtual pioerr init(const pio_pin* pin) = 0;
	
	//! \brief init a list of pins
	//! \param pins list of pins to init
	//! \param num num of pins in list
	//! \return @ref pioerr
	virtual pioerr init(const pio_pin** pins, uint32_t num) = 0;
	
	//! \brief deinit a single pin
	//! \param pin pin to deinit
	//! \return @ref pioerr
	virtual pioerr deinit(const pio_pin* pin) = 0;
	
	//! \brief deinit a list of pins
	//! \param depins list of pins to init
	//! \param num num of pins in list
	//! \return @ref pioerr
	virtual pioerr deinit(const pio_pin** pins, uint32_t num) = 0;
	
	//! \brief sets a single pin
	//! \param pin pin to set
	//! \return @ref pioerr
	virtual pioerr set(const pio_pin* pin) = 0;
	
	//! \brief clears a single pin
	//! \param pin pin to clear
	//! \return @ref pioerr
	virtual pioerr clear(const pio_pin* pin) = 0;
	
	//! \brief reads a single pin
	//! \param pin pin to read
	//! \param read pointer to store read status to (0 -> low; !0 -> high)
	//! \return @ref pioerr
	virtual pioerr getpin(const pio_pin* pin, uint32_t *read) = 0;
};

class pio_set {
public:

	pio_set();
	
	//! \brief init of this instance
	//! \param[in] pio_inst pio instance to use
	//! \param[in] config to use
	//! \return @ref pioerr
	pioerr init(i_pio* pio_inst, const pio_pin_set* config);
	
	//! \brief enables device
	//! \return @ref pioerr
	pioerr enable(void);
	
	//! \brief disables device
	//! \return @ref pioerr
	pioerr disable(void);

private:

	i_pio* m_pio_inst;

	const pio_pin_set* m_config;
	
};

#endif//#define I_PIO_HPP