/* 
 * This file is part of the Modbus IO distribution.
 * Copyright (c) 2020 Martin Wende.
 * 
 * This program is free software: you can redistribute it and/or modify  
 * it under the terms of the GNU General Public License as published by  
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "mb_uart32_de_irq.hpp"

#include <cstddef>

mb_uart32_de_irq::mb_uart32_de_irq() :
	m_config(NULL),
	m_callback_rx(NULL),
	m_callback_tx(NULL),
	m_userarg_rx(NULL),
	m_userarg_tx(NULL),
	m_reveive_buf(NULL),
	m_send_buf(NULL) 
{
 
}

mb_err mb_uart32_de_irq::init(const mb_uart_config *config) 
{
	m_config = NULL;
	if ((NULL == config) || (NULL == config->timer_inst))
	{
		return mb_err::ERR_ARGS;
	}
	
	timer_irq_err t_err;
	t_err = config->timer_inst->set_callback(timer_callb, this);
	if (timer_irq_err::SUCCESS != t_err)
	{
		return mb_err::ERR_INIT;
	}
	
	m_config = config;
	return mb_err::SUCCESS;
}

mb_err mb_uart32_de_irq::init(mb_baudrate baudrate, mb_parmode parity) 
{
	if (NULL == m_config)
	{
		return mb_err::ERR_INIT;
	}
	
	// check first if baudrate is possible
	mb_err err;
	err = check_baud(baudrate);
	if (mb_err::SUCCESS != err)
	{
		return err;
	}
	
	USART_TypeDef *usart = m_config->usart;
	
	// disable everything
	nvic_disable_irq(m_config->irq_nbr);
	usart->CR1 &= ~USART_CR1_UE; // some bits are only settable if UE = 0
	usart->CR1 = 0;
	usart->CR2 = 0;
	usart->CR3 = 0;
	
	usart->CR1 |= (Driver_Enable_assertion_time << USART_CR1_DEAT_Pos) & USART_CR1_DEAT_Msk;
	usart->CR1 |= (Driver_Enable_deassertion_time << USART_CR1_DEDT_Pos) & USART_CR1_DEDT_Msk;
	usart->CR1 |= USART_CR1_TE | USART_CR1_RE;
	
	usart->CR3 |= USART_CR3_DEM;
	
	uint32_t index = (uint32_t)baudrate;
	uint32_t baudrate_hz = baudlut[index] * 100;
	usart->BRR  = 16000000/baudrate_hz; // TODO: RCC
	
	// If a silent interval of more than 1.5 character times occurs between two characters, 
	// the message frame is declared incomplete and should be discarded by the receiver.
	
	// 3 chars timeout for "this was the last char"
	uint32_t time_us;
	uint32_t char_len = 1 + 8 + 1 + 1; // startbit, databyte, paritybit, stopbit (or 2 stopbit wo parity)
	
	// time per char
	#define SEK_TO_USEK 1000000 
	time_us  = (char_len*SEK_TO_USEK)/baudrate_hz;
	time_us *= 3;
	
	timer_irq_err t_err;
	t_err = m_config->timer_inst->set_time(time_us);
	if (timer_irq_err::SUCCESS != t_err)
	{
		return mb_err::ERR_INIT;
	}
	
	switch (parity) 
	{	
		case mb_parmode::EVENP_1S: // and default (else: error: duplicate case value)
			usart->CR1 |= USART_CR1_PCE;
			break;
		case mb_parmode::ODDP_1S: 
			usart->CR1 |= USART_CR1_PCE | USART_CR1_PS;
			break;
		case mb_parmode::NOP_2S:
			usart->CR2 |= 2 << USART_CR2_STOP_Pos;
			break;
		case mb_parmode::pmax:
			return mb_err::ERR_INIT;
	}
	
	nvic_enable_irq(m_config->irq_nbr, 0x02, 0x00);
	
	// enable at last
	usart->CR1 |= USART_CR1_UE;
	
	return mb_err::SUCCESS;
}

void mb_uart32_de_irq::deinit(void)
{
	if (NULL == m_config)
	{
		return ;
	}
	
	nvic_disable_irq(m_config->irq_nbr);
	
	if (m_config->send_en)
	{
		m_config->send_en->disable();
	}
	
	USART_TypeDef *usart = m_config->usart;
	usart->CR1 &= ~USART_CR1_UE;
	usart->CR1  = 0;
	m_config    = NULL;
}

mb_err mb_uart32_de_irq::check_baud(mb_baudrate baudrate) 
{
	const uint32_t send_deviation_promille = 10;
	uint32_t index = (uint32_t)baudrate;
	uint32_t baudrate_hz = baudlut[index] * 100;
	uint32_t bauddiv = 16000000/baudrate_hz; 	// TODO: RCC
	uint32_t baud_calcback = 16000000/bauddiv;	// TODO: RCC
	
	int32_t baud_diff_max  = baudrate_hz * send_deviation_promille;
	        baud_diff_max /= 1000;
	int32_t baud_diff = baudrate_hz - baud_calcback;
	if (baud_diff < 0)
	{
		baud_diff = 0 - baud_diff;
	}
	
	if (baud_diff > baud_diff_max)
	{
		return mb_err::ERR_BAUDRATE;
	}
	return mb_err::SUCCESS;
}

mb_err mb_uart32_de_irq::receive(uint8_t *buf, uint16_t len) 
{
	if (NULL == m_config)
	{
		return mb_err::ERR_INIT;
	}
	else if (NULL != m_reveive_buf)
	{
		return mb_err::ERR_BUSY;
	}
	
	m_receive_idx    = 0;
	m_receivelen_max = len;
	m_reveive_buf    = buf;
	
	// clear err flags and set irqs
	__disable_irq();
	USART_TypeDef *usart = m_config->usart;
	usart->ICR  = USART_ICR_PECF | USART_ICR_ORECF | USART_ICR_FECF;
	usart->CR1 |= USART_CR1_RXNEIE; // rx not empty irq
	__enable_irq();
	
	return mb_err::SUCCESS;
}

mb_err mb_uart32_de_irq::send(uint8_t *buf, uint16_t len) 
{
	if (NULL == m_config)
	{
		return mb_err::ERR_INIT;
	}
	else if (NULL != m_send_buf)
	{
		return mb_err::ERR_BUSY;
	}
	
	m_send_idx = 0;
	m_send_buf = buf;
	m_sendlen  = len;
	
	if (m_config->send_en)
	{
		m_config->send_en->enable();
	}
	
	// clear err flags and set irqs
	__disable_irq();
	USART_TypeDef *usart = m_config->usart;
	usart->ICR  = USART_ICR_TCCF;
	usart->CR1 |= USART_CR1_TXEIE; // tx not empty irq
	__enable_irq();
	
	return mb_err::SUCCESS;
}

mb_err mb_uart32_de_irq::register_callback_rx(mb_uart_callb callb, void *userarg) 
{
	m_callback_rx = callb;
	m_userarg_rx  = userarg;
	return mb_err::SUCCESS;
}

mb_err mb_uart32_de_irq::register_callback_tx(mb_uart_callb callb, void *userarg) 
{
	m_callback_tx = callb;
	m_userarg_tx  = userarg;
	return mb_err::SUCCESS;
}

void mb_uart32_de_irq::irq() 
{
	USART_TypeDef *usart = m_config->usart;
	uint32_t cr1 = usart->CR1;
	uint32_t sr  = usart->ISR;
	
	// send more
	if ((cr1 & USART_CR1_TXEIE) && (sr & USART_ISR_TXE))
	{
		usart->TDR = m_send_buf[m_send_idx++];
		
		if (m_send_idx == m_sendlen)
		{
			// switch to transmit clear
			usart->CR1 &= ~USART_CR1_TXEIE;
			usart->CR1 |=  USART_CR1_TCIE;
		}
	}
	
	// send last done
	if ((cr1 & USART_CR1_TCIE) && (sr & USART_ISR_TC))
	{
		uint8_t* buf = m_send_buf;
		m_send_buf   = NULL;
		
		usart->CR1 &= ~USART_CR1_TCIE;
		usart->ICR  =  USART_ICR_TCCF;
		
		if (m_config->send_en)
		{
			m_config->send_en->disable();
		}
		
		if (m_callback_tx)
		{
			m_callback_tx(buf, m_sendlen, 0, m_userarg_tx);
		}
	}
	
	// only react to RX if enabled
	if ((cr1 & USART_CR1_RXNEIE) && (sr & USART_ISR_RXNE))
	{
		m_config->timer_inst->restart();
		
		// are we getting too big?
		if (m_receivelen_max == m_receive_idx)
		{
			timer_callb(this);
		}
		else
		{
			m_reveive_buf[m_receive_idx++] = usart->RDR;
		}
	}
	
	nvic_pending_clear(m_config->irq_nbr);
}

void mb_uart32_de_irq::timer_callb(void* arg)
{
	if (NULL == arg)
	{
		return;
	}
	
	mb_uart32_de_irq *inst = (mb_uart32_de_irq*)arg;
	USART_TypeDef *usart = inst->m_config->usart;
	
	inst->m_config->timer_inst->stop();
	
	usart->CR1 &= ~USART_CR1_RXNEIE;
	uint32_t sr = usart->ISR;
	
	// uart errors set?
	uint32_t err = (uint32_t)mb_uart_err::SUCCESS;
	if (sr & USART_ISR_ORE)
	{
		err |= (uint32_t)mb_uart_err::ERR_OVERRUN;
		usart->ICR = USART_ICR_ORECF;
	}
	if (sr & USART_ISR_FE)
	{
		err |= (uint32_t)mb_uart_err::ERR_FRAME;
		usart->ICR = USART_ICR_FECF;
	}
	if (sr & USART_ISR_PE)
	{
		err |= (uint32_t)mb_uart_err::ERR_PARITY;
		usart->ICR = USART_ICR_PECF;
	}
	
	uint8_t *buf = inst->m_reveive_buf;
	inst->m_reveive_buf = NULL;
	
	if (inst->m_callback_rx)
	{
		inst->m_callback_rx(buf, inst->m_receive_idx, err, inst->m_userarg_rx);
	}
}