/* 
 * This file is part of the Modbus IO distribution.
 * Copyright (c) 2020 Martin Wende.
 * 
 * This program is free software: you can redistribute it and/or modify  
 * it under the terms of the GNU General Public License as published by  
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*
This is the implementation for STM32 with USART32 with Data Enable and Timeout
*/

#ifndef MB_UART32_DE_IRQ_HPP
#define MB_UART32_DE_IRQ_HPP

#include "i_mb_uart.hpp"

#include "cmsis_includer.h"
#include "i_timer_irq.hpp"
#include "i_pio.hpp"
#include "armcore/nvic.h"

#include <cstdint>

// in 1/16th sample time 0 - 15
#define Driver_Enable_assertion_time	8
#define Driver_Enable_deassertion_time	8

//! \brief modbus uart init struct
struct mb_uart_config {
	USART_TypeDef *usart;	   //!< uart instance
	IRQn_Type     irq_nbr; 	   //!< IRQ number on NVIC
	i_timer_irq   *timer_inst; //!< timer for rx timeouts
	pio_set*      send_en;     //!< opzional send enable pin if DE is not avail (NULL for disable)
};

class mb_uart32_de_irq : public i_mb_uart {
public:

	mb_uart32_de_irq();

	//! \brief init the config stuct in uart, but not the peripheral
	//! \param[in] dma_inst instance of dma to use
	//! \param[in] config config of uart
	//! \return @ref mb_err
	virtual mb_err init(const mb_uart_config *config);
	
	//! \brief init with uart mode
	//! \param[in] baudrate baudrate to set
	//! \param[in] parity parity to set
	//! \return @ref mb_err
	virtual mb_err init(mb_baudrate baudrate, mb_parmode parity);
	
	void deinit(void);
	
	//! \brief check if baudrate is possible before set
	//! \param[in] baudrate baudrate to set
	//! \return @ref mb_err
	virtual mb_err check_baud(mb_baudrate baudrate);
	
	//! \brief start receiving (again)
	//! \param[out] buf buffer to store received frame to (length is known)
	//! \param[in] len max length of frame to receive
	//! \return @ref mb_err
	virtual mb_err receive(uint8_t *buf, uint16_t len);
	
	//! \brief start sending
	//! \param[in] buf buffer to send
	//! \param[in] len length of frame to send
	//! \return @ref mb_err
	virtual mb_err send(uint8_t *buf, uint16_t len);
	
	//! \brief register a callback for receiving
	//! \param[in] callb callback to register (can be NULL)
	//! \param[in] userarg arg to call the callback with
	//! \return @ref mb_err
	virtual mb_err register_callback_rx(mb_uart_callb callb, void *userarg);
	
	//! \brief register a callback for sending
	//! \param[in] callb callback to register (can be NULL)
	//! \param[in] userarg arg to call the callback with
	//! \return @ref mb_err
	virtual mb_err register_callback_tx(mb_uart_callb callb, void *userarg);
	
	//! \brief call this if the UART IRQ happened
	virtual void irq();
	
private:

	//! \brief callback called by timer
	static void timer_callb(void* arg);
	
	//! \brief config of this uart
	const mb_uart_config *m_config;
	
	//! \brief maximum lenght to receive
	uint16_t m_receivelen_max;
	
	//! \brief idx in receive buffer
	uint16_t m_receive_idx;
	
	//! \brief pointer to buffer to receive
	uint8_t *m_reveive_buf;
	
	//! \brief length to send
	uint16_t m_sendlen;
	
	//! \brief idx in send buffer
	uint16_t m_send_idx;
	
	//! \brief pointer to buffer to send
	uint8_t *m_send_buf;
	
	//! \brief callbacks to upper layer 
	mb_uart_callb m_callback_rx;
	mb_uart_callb m_callback_tx;
	void *m_userarg_rx;
	void *m_userarg_tx;
};

#endif // MB_UART32_DE_IRQ_HPP