#include "pio.hpp"

#include <cstddef>

// 4 Bits per Alternate Function
#define AFR_BITS 4U
#define AFR_MSK  0xFU

// 2 Bits per Function
#define DUALBIT_MSK 0b11U
#define BITSPERPIN  2U

// coding of mode Register from enum PioPinMode
#define ENUM_REGMODER_MSK 0b11000U
#define ENUM_REGMODER_POS 3U
// coding of MODE Bits
#define ENUM_BITMODER_MSK 0b11U
#define ENUM_BITMODER_POS 0U
// coding of slewrate
#define ENUM_OTYPE_MSK 0b100U
#define ENUM_OTYPE_POS 2U
// coding of pull resistors
#define ENUM_PUPDR_MSK 0b11
#define ENUM_PUPDR_POS 0

// how many pins in one AFR Register
#define AFR_PINS_PER_REG 8

// 2 part Register BSRR for SET/RESET
#define BSSR_SET_POS 0U
#define BSSR_RES_POS 16U

pioerr pio::init(const pio_pin* pin){

	uint32_t pin_nbr = pin->nbr;
	uint32_t bitPosDualBit = pin->nbr * BITSPERPIN;
	if ((NULL == pin) || (pin_nbr >= 16U))
	{
		return pioerr::ERR_ARGS;
	}
	
	GPIO_TypeDef* gpio = pin->port;
	enum gpio_mode mode = pin->mode;
	
	// Altfunc set before MODER or else the wrong periph can have this pin
	volatile uint32_t *afr = &(gpio->AFR[0U]);
		
	// search for fitting AFR Register
	if (pin_nbr >= AFR_PINS_PER_REG){
		afr = &(gpio->AFR[1U]);
		pin_nbr -= AFR_PINS_PER_REG;
	}
		
	*afr &= ~(AFR_MSK << (pin_nbr*AFR_BITS));
	*afr |=  (pin->alt_func << (pin_nbr*AFR_BITS));
		
	// Mode (Input, GP Out, Alternate function, Analog)
	uint32_t regMode = (mode & ENUM_REGMODER_MSK) >> ENUM_REGMODER_POS;
	gpio->MODER &= ~(ENUM_BITMODER_MSK << (bitPosDualBit));
	gpio->MODER |= (regMode << (bitPosDualBit));
		
	// output type (Pushpull, Opendrain)
	uint32_t regPpod = (mode & ENUM_OTYPE_MSK) >> ENUM_OTYPE_POS;
	gpio->OTYPER &= ~(1 << pin_nbr);
	gpio->OTYPER |= (regPpod << pin_nbr);
		
	// Slewrate (Lowspeed, Mediumspeed, Fastspeed, Highspeed)
	gpio->OSPEEDR &= ~(DUALBIT_MSK << (bitPosDualBit));
	gpio->OSPEEDR |= (pin->speed << (bitPosDualBit));
		
	// Pullup, Pulldown
	uint32_t regPull = mode & ENUM_PUPDR_MSK;
	gpio->PUPDR &= ~(ENUM_PUPDR_MSK << (bitPosDualBit));
	gpio->PUPDR |= (regPull << (bitPosDualBit));
	
	return pioerr::SUCCESS;
}

pioerr pio::init(const pio_pin** pins, uint32_t num) {
	
	uint32_t i;
	for (i = 0; i < num ; i++) {
		init(pins[i]);
	}
	
	return pioerr::SUCCESS;
}

pioerr pio::deinit(const pio_pin* pin)
{
	if (NULL == pin)
	{
		return pioerr::ERR_ARGS;
	}
	
	pio_pin depin = {
		.port     = pin->port, 
		.nbr      = pin->nbr, 
		.mode     = GP_ANALOG, // default mode at bootup
		.speed    = GPIO_LS,
		.alt_func = 0,
	};
	
	return init(&depin);
}
	
pioerr pio::deinit(const pio_pin** pins, uint32_t num)
{
	uint32_t i;
	for (i = 0; i < num ; i++) {
		deinit(pins[i]);
	}
	
	return pioerr::SUCCESS;
}

pioerr pio::set(const pio_pin* pin){
	pin->port->BSRR = 1U << (pin->nbr + BSSR_SET_POS);
	return pioerr::SUCCESS;
}

pioerr pio::clear(const pio_pin* pin){
    pin->port->BSRR = 1U << (pin->nbr + BSSR_RES_POS);
	return pioerr::SUCCESS;
}

pioerr pio::getpin(const pio_pin* pin, uint32_t *read){
    *read = pin->port->IDR & (1U << pin->nbr);
	return pioerr::SUCCESS;
}