#ifndef PIO_HPP
#define PIO_HPP

#include "i_pio.hpp"

#include <stdint.h>

#include "cmsis_includer.h"

//! \brief enum of possible Pin Modes
enum gpio_mode {
	GPO_PP =		0b01000, //!< out pushpull
	GPO_PP_PU =		0b01001, //!< out pushpull + pullup
	GPO_PP_PD =		0b01010, //!< out pushpull + pulldown
	GPO_OD =		0b01100, //!< out opendrain
	GPO_OD_PU =		0b01101, //!< out opendrain + pullup
	GPO_OD_PD =		0b01110, //!< out opendrain + pulldown
	AF_PP =			0b10000, //!< altfunc pushpull
	AF_PP_PU =		0b10001, //!< altfunc pushpull + pullup
	AF_PP_PD =		0b10010, //!< altfunc pushpull + pulldown
	AF_OD =			0b10100, //!< altfunc opendrain	
	AF_OD_PU =		0b10101, //!< altfunc opendrain + pullup
	AF_OD_PD =		0b10110, //!< altfunc opendrain + pulldown
	GPI_FLOAT =		0b00000, //!< in floating
	GPI_PU =		0b00001, //!< in + pullup
	GPI_PD =		0b00010, //!< in + pulldown
	GP_ANALOG =		0b11000, //!< analog -> input is switched off
};

//! \brief enum of possible Pin Speeds
enum gpio_speed {
	GPIO_LS = 0b00, //!< low speed
	GPIO_MS = 0b01, //!< medium speed
	GPIO_FS = 0b10, //!< full speed
	GPIO_HS = 0b11, //!< high speed
};

//! \brief structure for one PIO pin 
struct pio_pin {
	GPIO_TypeDef* port;		//!< port of pin
	uint8_t nbr;			//!< pin Number (0 - 15)
	enum gpio_mode mode;	//!< pin mode @ref gpio_pin_mode
	enum gpio_speed speed;	//!< pin speed @ref PioPinSpeed
	uint8_t alt_func;		//!< alternate function, see datasheet
};

class pio : public i_pio {
public:
	
	//! \brief init a single pin
	//! \param pin pin to init
	//! \return @ref pioerr
	virtual pioerr init(const pio_pin* pin);
	
	//! \brief init a list of pins
	//! \param pins list of pins to init
	//! \param num num of pins in list
	//! \return @ref pioerr
	virtual pioerr init(const pio_pin** pins, uint32_t num);
	
	//! \brief deinit a single pin
	//! \param pin pin to deinit
	//! \return @ref pioerr
	virtual pioerr deinit(const pio_pin* pin);
	
	//! \brief deinit a list of pins
	//! \param depins list of pins to init
	//! \param num num of pins in list
	//! \return @ref pioerr
	virtual pioerr deinit(const pio_pin** pins, uint32_t num);
	
	//! \brief sets a single pin
	//! \param pin pin to set
	//! \return @ref pioerr
	virtual pioerr set(const pio_pin* pin);
	
	//! \brief clears a single pin
	//! \param pin pin to clear
	//! \return @ref pioerr
	virtual pioerr clear(const pio_pin* pin);
	
	//! \brief reads a single pin
	//! \param pin pin to read
	//! \param read pointer to store read status to (0 -> low; !0 -> high)
	//! \return @ref pioerr
	virtual pioerr getpin(const pio_pin* pin, uint32_t *read);
};

#endif//PIO_HPP