#include "timer_irq_base.hpp"

#include <cstddef>

timer_irq_base::timer_irq_base() :
	m_config(NULL),
	m_conf_sleep(NULL),
	m_debug_pin(NULL),
	m_callb(NULL),
	m_callb_arg(NULL)
{
}

timer_irq_err timer_irq_base::init(const timer_irq_base_config* config, timer_irq_base_sleep* conf_sleep, pio_set* debug_pin)
{
	m_config = NULL;
	
	if ((NULL == config) || (NULL == config->timer))
	{
		return timer_irq_err::ERR_ARGS;
	}
	
	m_config     = config;
	m_conf_sleep = conf_sleep;
	m_debug_pin  = debug_pin;
	
	TIM_TypeDef *tim = m_config->timer;
	tim->CR1   = 0;
	
	// downcount, one pulse, only underflow throws irq
	tim->CR1   = TIM_CR1_DIR | TIM_CR1_OPM | TIM_CR1_URS; 
	
	tim->CR2   = 0;
	tim->SMCR  = 0;
	tim->SR    = 0;
	tim->DIER |= TIM_DIER_UIE;
	
	nvic_enable_irq(m_config->irq_nbr, 0x02, 0x00);
	
	// we can sleep
	if (m_conf_sleep)
	{
		*m_conf_sleep->tim_sl_reg &= ~m_conf_sleep->tim_sl_bit;
	}
	
	return timer_irq_err::SUCCESS;
}

void timer_irq_base::deinit(void)
{
	if (NULL == m_config)
	{
		return;
	}
	
	nvic_disable_irq(m_config->irq_nbr);
	TIM_TypeDef *tim = m_config->timer;
	tim->CR1 &= ~TIM_CR1_CEN;
	tim->CR1  = 0;
	m_config  = NULL;
}
	
timer_irq_err timer_irq_base::set_time(uint32_t us)
{
	if (NULL == m_config)
	{
		return timer_irq_err::ERR_INIT;
	}
	else if (0 == us)
	{
		return timer_irq_err::ERR_ARGS;
	}
	
	nvic_disable_irq(m_config->irq_nbr);
	
	TIM_TypeDef *tim = m_config->timer;
	tim->CR1 &= ~TIM_CR1_CEN;
	
	// TODO: RCC
	uint64_t timer_clk    = 16000000;
	uint64_t timer_scale  = ((uint64_t)us*timer_clk);
	         timer_scale /= 1000000;
	         timer_scale += 1;
	
	uint16_t timer_psc    = timer_scale >> 16;
	uint16_t timer_arr    = timer_scale;
	
	tim->PSC = timer_psc;
	tim->ARR = timer_arr;
	
	// setting UIF for updating shadow registers
	tim->EGR = TIM_EGR_UG;
	tim->SR  = 0;
	
	nvic_enable_irq(m_config->irq_nbr, 0x02, 0x00);
	
	return timer_irq_err::SUCCESS;
}
	
timer_irq_err timer_irq_base::restart()
{
	if (NULL == m_config)
	{
		return timer_irq_err::ERR_INIT;
	}
	
	// no sleep anymore
	if (m_conf_sleep)
	{
		*m_conf_sleep->tim_sl_reg |= m_conf_sleep->tim_sl_bit;
	}
	
	
	TIM_TypeDef *tim = m_config->timer;
	tim->CNT   = tim->ARR;
	tim->SR    = 0;
	//tim->DIER |= TIM_DIER_UIE;
	tim->CR1  |= TIM_CR1_CEN;
	
	if (m_debug_pin)
	{
		m_debug_pin->enable();
	}
	
	return timer_irq_err::SUCCESS;
}
	
timer_irq_err timer_irq_base::stop()
{
	if (NULL == m_config)
	{
		return timer_irq_err::ERR_INIT;
	}
	
	TIM_TypeDef *tim = m_config->timer;
	tim->CR1  &= ~TIM_CR1_CEN;
	
	// we can sleep
	if (m_conf_sleep)
	{
		*m_conf_sleep->tim_sl_reg &= ~m_conf_sleep->tim_sl_bit;
	}
	
	return timer_irq_err::SUCCESS;
}
	
timer_irq_err timer_irq_base::set_callback(timer_irq_callb callb, void* arg)
{
	if (NULL == m_config)
	{
		return timer_irq_err::ERR_INIT;
	}
	
	m_callb     = callb;
	m_callb_arg = arg;
	
	return timer_irq_err::SUCCESS;
}
	
void timer_irq_base::irq()
{
	if (NULL == m_config)
	{
		return;
	}
	
	// we can sleep
	if (m_conf_sleep)
	{
		*m_conf_sleep->tim_sl_reg &= ~m_conf_sleep->tim_sl_bit;
	}
	
	TIM_TypeDef *tim = m_config->timer;
	tim->SR = 0;
	
	if (m_debug_pin)
	{
		m_debug_pin->disable();
	}
	
	if (m_callb)
	{
		m_callb(m_callb_arg);
	}
	
	nvic_pending_clear(m_config->irq_nbr);
}