#ifndef TIMER_IRQ_BASE_HPP
#define TIMER_IRQ_BASE_HPP

#include <cstdint>

#include "cmsis_includer.h"
#include "i_pio.hpp"
#include "i_timer_irq.hpp"
#include "armcore/nvic.h"

struct timer_irq_base_config
{
	TIM_TypeDef *timer;	 //!< timer instance
	IRQn_Type irq_nbr;   //!< IRQ number on NVIC
};

//! \brief sleep config of timer
struct timer_irq_base_sleep {
	volatile uint32_t* tim_sl_reg; //!< Register to enable/disable sleep of timer
	         uint32_t  tim_sl_bit; //!< Bit      to enable/disable sleep of timer
};

class timer_irq_base : public i_timer_irq {
public:

	timer_irq_base();
	
	//! \brief init timer irq instance
	//! \param config config for timer
	//! \param conf_sleep optional (can be NULL) sleep config of this timer
	//! \param debug_pin debug pin for seeing timings (H for timer duration), (NULL disables)
	//! \return @ref timer_irq_err
	virtual timer_irq_err init(const timer_irq_base_config* config, timer_irq_base_sleep* conf_sleep, pio_set* debug_pin);
	
	void deinit(void);
	
	//! \brief init a list of pins
	//! \param us to set time in us
	//! \return @ref timer_irq_err
	virtual timer_irq_err set_time(uint32_t us);
	
	//! \brief (re)starts a timer
	//! \return @ref timer_irq_err
	virtual timer_irq_err restart();
	
	//! \brief stops a timer
	//! \return @ref timer_irq_err
	virtual timer_irq_err stop();
	
	//! \brief reads a single pin
	//! \param callb callback to set (NULL for deleting)
	//! \param arg arg given to callback (NULL allowed)
	//! \return @ref timer_irq_err
	virtual timer_irq_err set_callback(timer_irq_callb callb, void* arg);
	
	//! \brief call this if the Timer IRQ happened
	void irq();
	
private:
	
	//! \brief config for this instance
	const timer_irq_base_config* m_config;
	
	//! \brief optional sleep conf
	timer_irq_base_sleep* m_conf_sleep;
	
	pio_set* m_debug_pin;
	
	//! \brief callback to call
	timer_irq_callb m_callb;
	
	//! \brief argument for callback
	void* m_callb_arg;
};

#endif // TIMER_IRQ_BASE_HPP